# Add manager to user
add_manager_template = """
Get-ADUser -Filter 'SamAccountName -eq \"{user_sam}\"' -Credential $usercredential |
Set-ADUser -Manager {manager_sam} -Credential $usercredential
"""


rename_user_template = """
Get-ADUser -Filter 'SamAccountName -eq \"{user_sam}\"' -Credential $usercredential |
Rename-ADobject -Credential $usercredential -NewName "{new_name}"
"""



import datetime
import json
import logging
import sys
from functools import partial
from itertools import chain
from typing import Any
from typing import Optional
from typing import List
from typing import OrderedDict

import click
import requests
from os2mo_helpers.mora_helpers import MoraHelper

from .date_utils import (
    datetime_to_sd_date,
    parse_datetime,
    SD_INFINITY,
    MO_INFINITY,
    format_date,
)
from . import sd_payloads
from .config import ChangedAtSettings
from .config import get_changed_at_settings
from .exceptions import NoCurrentValdityException
from .sd_common import mora_assert
from .sd_common import sd_lookup


LOG_LEVEL = logging.DEBUG

logger = logging.getLogger("fixDepartments")


def setup_logging():
    detail_logging = ("sdCommon", "fixDepartments")
    for name in logging.root.manager.loggerDict:
        if name in detail_logging:
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        stream=sys.stdout,
    )


class FixDepartments:
    def __init__(self, settings: ChangedAtSettings, dry_run: bool = False):
        logger.info("Start program")
        self.settings = settings
        self.dry_run = dry_run

        self.institution_uuid = self.get_institution()
        self.helper = self._get_mora_helper(self.settings)

        if self.settings.sd_fix_departments_root is not None:
            self.org_uuid = str(settings.sd_fix_departments_root)
        else:
            try:
                self.org_uuid = self.helper.read_organisation()
            except requests.exceptions.RequestException as e:
                logger.error(e)
                print(e)
                exit()
            except json.decoder.JSONDecodeError as e:
                logger.error(e)
                print(e)
                exit()

        logger.info("Read org_unit types")
        self.level_types = self.helper.read_classes_in_facet("org_unit_level")[0]
        unit_types = self.helper.read_classes_in_facet("org_unit_type")[0]

        # Currently only a single unit type exists, we will not do anything fancy
        # until it has been decided what the source for types should be.
        self.unit_type = None
        for unit in unit_types:
            if unit["user_key"] == "Enhed":
                self.unit_type = unit

        if self.unit_type is None:
            raise Exception("Unit types not correctly configured")

    def _get_mora_helper(self, settings):
        return MoraHelper(hostname=self.settings.mora_base, use_cache=False)

    def get_institution(self):
        """
        Get the institution uuid of the current organisation. It is uniquely
        determined from the InstitutionIdentifier. The identifier is read
        from settings.json. The value is rarely used, but is needed to dertermine
        if a unit is a root unit.
        :return: The SD institution uuid for the organisation.
        """
        inst_id = self.settings.sd_institution_identifier
        params = {"UUIDIndicator": "true", "InstitutionIdentifier": inst_id}
        institution_info = sd_lookup(
            "GetInstitution20111201", settings=self.settings, params=params
        )
        institution = institution_info["Region"]["Institution"]
        institution_uuid = institution["InstitutionUUIDIdentifier"]
        return institution_uuid

    def create_single_department(
        self, department: OrderedDict, parent_uuid: str
    ) -> None:
        """
        Create an organization unit in MO based on the corresponding department info
        from SD.

        Args:
            department: The SD department
            parent_uuid: The parent UUID of the department in SD
        """
        logger.info(
            f"""Create department with SD UUID: {department["DepartmentUUIDIdentifier"]}"""
        )

        effective_parent_uuid = (
            parent_uuid if parent_uuid is not None else self.org_uuid
        )

        for unit_level in self.level_types:
            if unit_level["user_key"] == department["DepartmentLevelIdentifier"]:
                unit_level_uuid = unit_level["uuid"]

        logger.debug("SD department info to create from: {}".format(department))
        payload = sd_payloads.create_single_org_unit(
            department=department,
            unit_type=self.unit_type["uuid"],
            unit_level=unit_level_uuid,
            parent=effective_parent_uuid,
        )
        logger.debug("Create department MO payload: {}".format(payload))
        if self.dry_run:
            print("Dry-run (ou/create): ", payload)
            return None
        response = self.helper._mo_post("ou/create", payload)
        response.raise_for_status()
        logger.info("Created unit {}".format(department["DepartmentIdentifier"]))
        logger.debug("Create response status: {}".format(response.status_code))

    def _create_org_unit_if_missing_in_mo(
        self, department: OrderedDict, parent_uuid: Optional[str]
    ) -> bool:
        """
        Create SD department in MO if it is missing.

        Args:
            department: the SD department
            parent_uuid: UUID of the parent unit

        Returns:
            Boolean indicating whether the unit was created in MO or not.
        """
        # This is a hack which should be removed once
        # https://redmine.magenta-aps.dk/issues/56846 has been resolved
        date_at = parse_datetime(department["ActivationDate"])
        if date_at <= datetime.datetime(1930, 1, 1):
            date_at = datetime.datetime(1930, 1, 2)
        date_at_str = format_date(date_at)

        mo_response = self.helper.read_ou(
            department["DepartmentUUIDIdentifier"], at=date_at_str
        )

        ou_created = False
        if mo_response.get("status") == 404:
            # TODO: create_single_department should return the boolean value of ou_created
            self.create_single_department(department, parent_uuid)
            ou_created = True

        return ou_created

    def _update_org_unit_for_single_sd_dep_registration(
        self, department: OrderedDict, parent_uuid: Optional[str]
    ) -> None:
        # Get SD department data
        unit_uuid = department["DepartmentUUIDIdentifier"]
        name = department["DepartmentName"]
        shortname = department["DepartmentIdentifier"]
        department_level_identifier = department["DepartmentLevelIdentifier"]
        from_date = department["ActivationDate"]
        to_date = department["DeactivationDate"]
        to_date = MO_INFINITY if to_date == SD_INFINITY else to_date

        unit_level_uuid = None
        for unit_level in self.level_types:
            if unit_level["user_key"] == department_level_identifier:
                unit_level_uuid = unit_level["uuid"]
        if unit_level_uuid is None:
            msg = "Unknown department level {}!!"
            logger.error(msg.format(department_level_identifier))
            raise Exception(msg.format(department_level_identifier))

        parent = parent_uuid if parent_uuid is not None else self.org_uuid

        msg = "Unit parent at {} is {}"
        logger.info(msg.format(from_date, parent))

        payload = sd_payloads.edit_org_unit(
            user_key=shortname,
            name=name,
            unit_uuid=unit_uuid,
            parent=parent,
            ou_level=unit_level_uuid,
            ou_type=self.unit_type["uuid"],
            from_date=from_date,
            to_date=to_date,
        )
        logger.debug("Edit payload to fix unit: {}".format(payload))
        if self.dry_run:
            print("Dry-run (details/edit): ", payload)
            return None
        response = self.helper._mo_post("details/edit", payload)
        logger.debug("Edit response status: {}".format(response.status_code))
        if response.status_code == 400:
            assert response.text.find("raise to a new registration") > 0
        else:
            response.raise_for_status()

    def fix_department(self, unit_uuid: str, validity_date: datetime.date) -> None:
        """
        Synchronize the state of a MO unit to the current and future state(s) in SD.
        :param unit_uuid: uuid of the unit to be updated.
        :param validity_date: The validity date to read the department info from SD.
        """

        msg = "Set department {} to state as of {}"
        logger.info(msg.format(unit_uuid, validity_date))
        validity = {
            "from_date": validity_date.strftime("%d.%m.%Y"),
            "to_date": datetime_to_sd_date(parse_datetime(SD_INFINITY)),
        }

        departments = self.get_department(validity, uuid=unit_uuid)
        for department in departments:
            # Get the UUID of the parent unit
            parent_lookup_date = max(
                validity_date, parse_datetime(department["ActivationDate"]).date()
            )
            parent_uuid = self.get_parent(unit_uuid, parent_lookup_date)

            # Create org unit if missing in MO
            ou_created = self._create_org_unit_if_missing_in_mo(department, parent_uuid)

            # ... and fix the parent before updating the unit itself
            if parent_uuid is not None:
                self.fix_department(parent_uuid, validity_date)

            if not ou_created:
                self._update_org_unit_for_single_sd_dep_registration(
                    department, parent_uuid
                )

    def get_department(
        self, validity, shortname=None, uuid=None
    ) -> List[OrderedDict[str, Any]]:
        """
        Read department information from SD.
        NOTICE: Shortnames are not universally unique in SD, and even a request
        spanning a single date might return more than one row if searched by
        shortname.
        :param validity: Validity dictionaty.
        :param shortname: Shortname for the unit(s).
        :param uuid: uuid for the unit.
        :return: A list of information about the unit(s).
        """
        params = {
            "ActivationDate": validity["from_date"],
            "DeactivationDate": validity["to_date"],
            "ContactInformationIndicator": "true",
            "DepartmentNameIndicator": "true",
            "PostalAddressIndicator": "false",
            "ProductionUnitIndicator": "false",
            "UUIDIndicator": "true",
            "EmploymentDepartmentIndicator": "false",
        }
        if uuid is not None:
            params["DepartmentUUIDIdentifier"] = uuid
        if shortname is not None:
            params["DepartmentIdentifier"] = shortname

        if uuid is None and shortname is None:
            raise Exception("Provide either uuid or shortname")

        department_info = sd_lookup(
            "GetDepartment20111201", settings=self.settings, params=params
        )
        department = department_info.get("Department")
        if department is None:
            raise NoCurrentValdityException()
        if isinstance(department, OrderedDict):
            department = [department]
        return department

    # Notice! Similar code also exists in sd_changed_at
    def _find_engagement(self, mo_engagements, job_id):
        """
        Given a list of engagements for a person, find the one with a specific
        job_id. If several elements covering the same engagement is in the list
        an unspecified element will be returned.
        :param mo_engaements: A list of engagements as returned by MO.
        :param job_id: The SD JobIdentifier to find.
        :return: Some element in the list that has the correct job_id. If no
        engagement is found, None is returned.
        """
        relevant_engagement = None
        try:
            user_key = str(int(job_id)).zfill(5)
        except ValueError:  # We will end here, if int(job_id) fails
            user_key = job_id

        for mo_eng in mo_engagements:
            if mo_eng["user_key"] == user_key:
                relevant_engagement = mo_eng

        if relevant_engagement is None:
            msg = "Fruitlessly searched for {} in {}".format(job_id, mo_engagements)
            logger.info(msg)
        return relevant_engagement

    def _read_department_engagements(self, unit_uuid, validity_date):
        """
        Retrive a list from SD with all engagements in a given department.
        All current (as of validity_date) and future engagements are retrived,
        since GetEngagement does not support time ranges, we ask for three
        points in time that should cover all known future.
        :param unit_uuid: uuid of the relevant department.
        :param validity_date: The origin of the query, all engagements newer than
        this will be retrieved.
        :return: Dict with cpr as key and SD Person objects as values.
        """
        fix_date = validity_date + datetime.timedelta(weeks=80)
        too_deep = self.settings.sd_import_too_deep
        sd_validity = {
            "from_date": fix_date.strftime("%d.%m.%Y"),
            "to_date": fix_date.strftime("%d.%m.%Y"),
        }
        department = self.get_department(sd_validity, uuid=unit_uuid)[0]
        if not department["DepartmentLevelIdentifier"] in too_deep:
            msg = "{} regnes ikke som et SD afdelingsniveau"
            logger.info(msg.format(unit_uuid))
            return {}

        params = {
            "DepartmentIdentifier": department["DepartmentIdentifier"],
            "DepartmentLevelIdentifier": department["DepartmentLevelIdentifier"],
            "StatusActiveIndicator": True,
            "StatusPassiveIndicator": False,
            "DepartmentIndicator": True,
            "UUIDIndicator": True,
        }

        # We need to catch all current and future engagements, this is an attempt to
        # do so, without making too many calls to the api.
        time_deltas = [0, 90, 365]

        all_people = {}
        logger.debug("Perform GetEmployments, time_delas: {}".format(time_deltas))
        for time_delta in time_deltas:
            effective_date = validity_date + datetime.timedelta(days=time_delta)
            params["EffectiveDate"] = (effective_date.strftime("%d.%m.%Y"),)

            employments = sd_lookup(
                "GetEmployment20111201", settings=self.settings, params=params
            )
            people = employments.get("Person", [])
            if not isinstance(people, list):
                people = [people]

            for person in people:
                cpr = person["PersonCivilRegistrationIdentifier"]
                if cpr not in all_people:
                    all_people[cpr] = person
        logger.debug("Department engagements: {}".format(all_people.keys()))
        return all_people

    def fix_NY_logic(self, unit_uuid, validity_date):
        """
        Read all engagements in a unit and ensure that the position in MO is correct
        according to the rules of the import (ie, move engagement up from
        'Afdeling'). This is mainly relevant if an 'Afdeling' is moved to have a new
        NY-department as parent.
        Notice that this should be called AFTER the recursive fix of the department
        tree to ensure that the logic is applied to a self-consistent tree.

        If an engagement is already correct, it will not be moved, if it is currently
        in a wrong unit, it will be corrected.
        :param unit_uuid: uuid of the unit to check.
        :validity_date: The validity_date of the operation, moved engagements will
        be moved as of this date.
        """
        too_deep = self.settings.sd_import_too_deep
        mo_unit = self.helper.read_ou(unit_uuid)
        while mo_unit["org_unit_level"]["user_key"] in too_deep:
            mo_unit = mo_unit["parent"]
            logger.debug("Parent unit: {}".format(mo_unit["uuid"]))
        destination_unit = mo_unit["uuid"]
        logger.debug("Destination found: {}".format(destination_unit))

        all_people = self._read_department_engagements(unit_uuid, validity_date)

        # We now have a list of all current and future people in the unit,
        # they should all be unconditionally moved if they are not already
        # in destination_unit.
        for person in all_people.values():
            cpr = person["PersonCivilRegistrationIdentifier"]

            if not isinstance(person["Employment"], list):
                person["Employment"] = [person["Employment"]]

            for employment in person["Employment"]:
                job_id = employment["EmploymentIdentifier"]
                msg = "Checking job-id: {}"
                print(msg.format(job_id))
                logger.info(msg.format(job_id))
                sd_uuid = employment["EmploymentDepartment"]["DepartmentUUIDIdentifier"]
                if not sd_uuid == unit_uuid:
                    # This employment is not from the current department,
                    # but is inherited from a lower level. Can happen if this
                    # tool is initiated on a level higher than Afdelings-niveau.
                    continue

                mo_person = self.helper.read_user(user_cpr=cpr, org_uuid=self.org_uuid)
                if mo_person is None:
                    msg = "MO person is None for job_id: {}"
                    logger.warning(msg.format(job_id))
                    continue

                mo_engagements = self.helper.read_user_engagement(
                    mo_person["uuid"], read_all=True, only_primary=True, skip_past=True
                )

                # Find the uuid of the relevant engagement and update all current and
                # future rows.
                mo_engagement = self._find_engagement(mo_engagements, job_id)
                if mo_engagement is None:
                    msg = "MO engagement is None for job_id: {}, user_uuid: {}"
                    logger.warning(msg.format(job_id, mo_person["uuid"]))
                    continue
                for eng in mo_engagements:
                    if not eng["uuid"] == mo_engagement["uuid"]:
                        # This engagement is not relevant for this unit
                        continue
                    if eng["org_unit"]["uuid"] == destination_unit:
                        # This engagement is already in the correct unit
                        continue

                    from_date: datetime.date = datetime.datetime.strptime(
                        eng["validity"]["from"], "%Y-%m-%d"
                    ).date()
                    if from_date < validity_date:
                        eng["validity"]["from"] = validity_date.strftime("%Y-%m-%d")

                    data = {
                        "org_unit": {"uuid": destination_unit},
                        "validity": eng["validity"],
                    }
                    payload = sd_payloads.engagement(data, mo_engagement)
                    logger.debug("Move engagement payload: {}".format(payload))
                    if self.dry_run:
                        print("Dry-run (details/edit): ", payload)
                        return
                    response = self.helper._mo_post("details/edit", payload)
                    mora_assert(response)

    def get_parent(self, unit_uuid, validity_date) -> Optional[str]:
        """
        Return the parent of a given department at at given point in time.
        Notice that the query is perfomed against SD, not against MO.
        It is generally not possible to predict whether this call will succeed, since
        this depends on the internal start-date at SD, which cannot be read from the
        API; the user of this function should be prepared to handle
        NoCurrentValdityException, unless the validity of the unit is known from
        other sources. In general queries to the future and near past should always
        be safe if the unit exists at the point in time.
        :param unit_uuid: uuid of the unit to be queried.
        :param validity_date: python datetime object with the date to query.
        :return: uuid of the parent department, None if the department is a root.
        """
        params = {
            "EffectiveDate": validity_date.strftime("%d.%m.%Y"),
            "DepartmentUUIDIdentifier": unit_uuid,
        }
        parent_response = sd_lookup(
            "GetDepartmentParent20190701", settings=self.settings, params=params
        )
        if "DepartmentParent" not in parent_response:
            msg = "No parent for {} found at validity: {}"
            logger.error(msg.format(unit_uuid, validity_date))
            raise NoCurrentValdityException()
        parent = parent_response["DepartmentParent"]["DepartmentUUIDIdentifier"]
        if parent == self.institution_uuid:
            parent = None
        return parent

    def get_all_parents(self, leaf_uuid, validity_date):
        """
        Find all parents from leaf unit up to the root of the tree.
        Notice, this is a query to SD, not to MO.
        :param leaf_uuid: The starting point of the chain, this does not stictly need
        to be a leaf node.
        :validity_date: The validity date of the fix.
        :return: A list of tuples containing short names and unit uuids sorted from leaf to root.
        """
        validity = {
            "from_date": validity_date.strftime("%d.%m.%Y"),
            "to_date": validity_date.strftime("%d.%m.%Y"),
        }
        department_branch = []
        department = self.get_department(validity=validity, uuid=leaf_uuid)[0]
        department_branch.append((department["DepartmentIdentifier"], leaf_uuid))

        current_uuid = self.get_parent(
            department["DepartmentUUIDIdentifier"], validity_date=validity_date
        )

        while current_uuid is not None:
            current_uuid = self.get_parent(
                department["DepartmentUUIDIdentifier"], validity_date=validity_date
            )
            department = self.get_department(validity=validity, uuid=current_uuid)[0]
            shortname = department["DepartmentIdentifier"]
            level = department["DepartmentLevelIdentifier"]
            uuid = department["DepartmentUUIDIdentifier"]
            department_branch.append((shortname, uuid))
            current_uuid = self.get_parent(current_uuid, validity_date=validity_date)
            msg = "Department: {}, uuid: {}, level: {}"
            logger.debug(msg.format(shortname, uuid, level))
        return department_branch

    def sd_uuid_from_short_code(self, validity_date, shortname):
        validity = {
            "from_date": validity_date.strftime("%d.%m.%Y"),
            "to_date": validity_date.strftime("%d.%m.%Y"),
        }
        department = self.get_department(validity, shortname=shortname)[0]
        return department["DepartmentUUIDIdentifier"]


@click.command()
@click.option(
    "--department-short-name",
    "short_names",
    multiple=True,
    type=click.STRING,
    help="Shortname of the department to update",
)
@click.option(
    "--department-uuid",
    "uuids",
    multiple=True,
    type=click.UUID,
    help="UUID of the department to update",
)
def unit_fixer(short_names, uuids):
    """Sync SD department information to MO."""
    setup_logging()

    settings = get_changed_at_settings()
    unit_fixer = FixDepartments(settings)

    today = datetime.datetime.today().date()

    # Translate short_names to uuids
    short_name_uuids = map(
        partial(unit_fixer.sd_uuid_from_short_code, today), short_names
    )
    # Convert UUIDs to strings
    department_uuids = map(str, uuids)

    for department_uuid in chain(short_name_uuids, department_uuids):
        unit_fixer.fix_department(department_uuid, today)
        unit_fixer.fix_NY_logic(department_uuid, today)


if __name__ == "__main__":
    unit_fixer()



import payloads
from hypothesis import given
from hypothesis import strategies as st


class TestConvertCreateToEdit:
    @given(st.dictionaries(keys=st.text(), values=st.text()), st.text(), st.text())
    def test_convert_no_from_date(self, payload, uuid_arg, type_arg):

        payload["uuid"] = uuid_arg
        payload["type"] = type_arg

        actual = payloads.convert_create_to_edit(payload)

        assert actual["data"] == payload
        assert actual["uuid"] == uuid_arg
        assert actual["type"] == type_arg

    @given(
        st.dictionaries(keys=st.text(), values=st.text()),
        st.text(),
        st.text(),
        st.text(min_size=1),
    )
    def test_convert_with_from_date(self, payload, uuid_arg, type_arg, from_date):

        payload["uuid"] = uuid_arg
        payload["type"] = type_arg

        actual = payloads.convert_create_to_edit(payload, from_date)

        assert actual["data"]["validity"]["from"] == from_date
        assert actual["uuid"] == uuid_arg
        assert actual["type"] == type_arg



import logging
import random
from operator import itemgetter

import click
import sentry_sdk
from click_option_group import optgroup
from click_option_group import RequiredMutuallyExclusiveOptionGroup
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.apply import apply
from ra_utils.load_settings import load_settings
from ra_utils.tqdm_wrapper import tqdm

from .ad_common import AD
from .ad_exceptions import ImproperlyConfigured
from .ad_logger import start_logging
from .ad_reader import ADParameterReader

logger = logging.getLogger("MoUuidAdSync")


class SyncMoUuidToAd(AD):
    """Syncronize MO UUIDs to AD.

    Walks through all users in AD, search in MO by cpr and writes the MO
    uuid on the users AD account.
    """

    def __init__(self):
        super().__init__()
        self.settings = load_settings()
        self._check_ad_uuid_field_is_configured()
        self.helper = self._get_mora_helper()
        self.org_uuid = self.helper.read_organisation()
        self.reader = ADParameterReader()
        self.stats = {
            "attempted_users": 0,
            "user_not_in_mo": 0,
            "already_ok": 0,
            "updated": 0,
        }

    def perform_sync(self, ad_users, mo_users):
        separator = self.all_settings["primary"].get("cpr_separator", "")
        cpr_field = self.all_settings["primary"]["cpr_field"]

        def extract_cpr(ad_user):
            self.stats["attempted_users"] += 1
            cpr = ad_user.get(cpr_field)
            if separator:
                cpr = cpr.replace(separator, "")
            return ad_user, cpr

        @apply
        def lookup_mo_uuid(ad_user, cpr):
            mo_uuid = mo_users.get(cpr)
            return ad_user, mo_uuid

        @apply
        def filter_unmatched_mo(ad_user, mo_uuid):
            if not mo_uuid:
                logger.info("cpr not in MO, sam {}".format(ad_user["SamAccountName"]))
                self.stats["user_not_in_mo"] += 1
                return False
            return True

        ad_uuid_field = self.settings["integrations.ad.write.uuid_field"]

        @apply
        def filter_already_synced(ad_user, mo_uuid):
            expected_mo_uuid = ad_user.get(ad_uuid_field)
            if expected_mo_uuid is None:
                msg = (
                    "Read None from UUID field in AD!"
                    + "Perhaps uuid_field or properties are misconfigured?"
                )
                logger.warning(msg)
            if expected_mo_uuid == mo_uuid:
                logger.info("uuid for {} correct in AD".format(mo_uuid))
                self.stats["already_ok"] += 1
                return False
            logger.debug(
                "uuid for {} not correct in AD: {}".format(mo_uuid, expected_mo_uuid)
            )
            return True

        server_strings = [""]
        if self.all_settings["global"].get("servers"):
            server_strings = [
                " -Server {} ".format(server)
                for server in self.all_settings["global"].get("servers")
            ]

        @apply
        def construct_powershell_script(ad_user, mo_uuid):
            logger.debug("Syncronizing uuid {} into AD".format(mo_uuid))
            server_string = random.choice(server_strings)
            ps_script = (
                self._build_user_credential()
                + "Get-ADUser "
                + server_string
                + " -Filter 'SamAccountName -eq \""
                + ad_user["SamAccountName"]
                + "\"' -Credential $usercredential | "
                + " Set-ADUser -Credential $usercredential "
                + ' -Replace @{"'
                + ad_uuid_field
                + '"="'
                + mo_uuid
                + '"} '
                + server_string
            )
            logger.debug("PS-script: {}".format(ps_script))
            return ps_script

        logger.info("Will now process {} users".format(len(ad_users)))

        users = tqdm(ad_users)
        users = map(extract_cpr, users)
        users = map(lookup_mo_uuid, users)
        users = filter(filter_unmatched_mo, users)
        users = filter(filter_already_synced, users)
        users = list(users)

        print(self.stats)
        logger.info(self.stats)

        logger.info("Will now attempt to sync {} users".format(len(users)))
        users = tqdm(users)
        users = map(construct_powershell_script, users)

        # Actually fire the powershell scripts, and trigger side-effects
        for ps_script in users:
            try:
                response = self._run_ps_script(ps_script)
            except Exception:
                logger.exception("failed to write MO UUID (ps_script=%r)", ps_script)
            else:
                logger.debug("Response: {}".format(response))
                if response:
                    msg = "Unexpected response: {}".format(response)
                    logger.exception(msg)
                    raise Exception(msg)
                self.stats["updated"] += 1
        print(self.stats)
        logger.info(self.stats)

    def sync_one(self, cprno):
        print("Fetch AD User")
        ad_user = self.reader.read_user(cpr=cprno)
        if not ad_user:
            msg = "AD User not found"
            logger.exception(msg)
            raise Exception(msg)

        print("Fetch MO User")
        mo_user = self.helper.read_user(user_cpr=cprno, org_uuid=self.org_uuid)
        try:
            mo_uuid = mo_user["uuid"]
        except (KeyError, TypeError):
            msg = "MO User not found"
            logger.exception(msg)
            raise Exception(msg)

        ad_users = [ad_user]
        mo_users = {cprno: mo_uuid}

        print("Starting Sync")
        self.perform_sync(ad_users, mo_users)

    def sync_all(self):
        def has_cpr_no(mo_user):
            if "cpr_no" not in mo_user:
                logger.warning("no 'cpr_no' for MO user %r", mo_user["uuid"])
                return False
            return True

        print("Fetch AD Users")
        ad_users = self.reader.read_it_all(print_progress=True)

        print("Fetch MO Users")
        mo_users = self.helper.read_all_users()
        mo_users = filter(has_cpr_no, mo_users)
        mo_users = dict(map(itemgetter("cpr_no", "uuid"), mo_users))

        print("Starting Sync")
        self.perform_sync(ad_users, mo_users)

    def _check_ad_uuid_field_is_configured(self):
        # Check configuration
        ad_uuid_field = self.settings["integrations.ad.write.uuid_field"]
        if not any(
            ad_uuid_field in ad_settings["properties"]
            for ad_settings in self.settings["integrations.ad"]
        ):
            msg = "'uuid_field' %r not in 'properties' of any AD"
            logger.warning(msg, ad_uuid_field)
            raise ImproperlyConfigured(msg % (ad_uuid_field))

    def _get_mora_helper(self):
        return MoraHelper(
            hostname=self.all_settings["global"]["mora.base"], use_cache=False
        )


@click.command()
@click.option(
    "--debug",
    help="Set logging level to DEBUG (default is INFO)",
    is_flag=True,
    default=False,
)
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option("--sync-all", is_flag=True)
@optgroup.option("--sync-cpr")
def cli(**args):
    start_logging(export_log_file="sync_mo_uuid_to_ad.log")

    # Set log level according to --debug command line arg
    logger.setLevel(logging.INFO)

    if args.get("debug"):
        logger.setLevel(logging.DEBUG)

    logger.debug(args)

    sync = SyncMoUuidToAd()

    if "crontab.SENTRY_DSN" in sync.settings:
        sentry_sdk.init(dsn=sync.settings["crontab.SENTRY_DSN"])

    if args.get("sync_all"):
        sync.sync_all()
    if args.get("sync_cpr"):
        sync.sync_one(args["sync_cpr"])
    logger.info("Sync done")


if __name__ == "__main__":
    cli()



import asyncio
from types import ModuleType
from typing import Optional
from typing import Tuple
from unittest import mock
from uuid import UUID

import config
import dar_helper
import initial
import los_files
import util


class HelperMixin:
    def _run_until_complete(self, coro):
        future = asyncio.ensure_future(coro)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(future)

    def _mock_dar_lookup(self, return_value: Tuple[Optional[str], Optional[UUID]]):
        return mock.patch.object(
            dar_helper,
            "dar_datavask_multiple",
            return_value=[return_value],
        )

    def _mock_read_csv(self, instance):
        return mock.patch.object(los_files, "read_csv", return_value=[instance])

    def _mock_util_call(self, name, return_value=None):
        return mock.patch.object(
            util, name, return_value=return_value or mock.MagicMock()
        )

    def _mock_get_client_session(self):
        return self._mock_util_call("get_client_session")

    def _mock_create_details(self):
        return self._mock_util_call("create_details")

    def _mock_edit_details(self):
        return self._mock_util_call("edit_details")

    def _mock_terminate_details(self):
        return self._mock_util_call("terminate_details")

    def _mock_lookup_employees(self, return_value=None):
        return self._mock_util_call("lookup_employees", return_value=return_value)

    def _mock_lookup_organisationfunktion(self, return_value=None):
        return self._mock_util_call(
            "lookup_organisationfunktion", return_value=return_value
        )

    def _mock_get_fileset_implementation(self, fileset):
        return mock.patch.object(
            los_files, "get_fileset_implementation", return_value=fileset
        )

    def _mock_get_import_filenames(self, filenames, created_date):
        fileset = mock.Mock(spec=los_files.FileSet)
        fileset.get_import_filenames = mock.Mock(return_value=filenames)
        fileset.get_modified_datetime = mock.Mock(return_value=created_date)
        return self._mock_get_fileset_implementation(fileset)

    def _mock_get_modified_datetime(self, return_value=None, side_effect=None):
        fileset = mock.Mock(spec=los_files.FileSet)
        fileset.get_modified_datetime = mock.Mock(
            return_value=return_value, side_effect=side_effect
        )
        return self._mock_get_fileset_implementation(fileset)

    def _mock_settings_json(self, settings=None):
        return mock.patch("config.load_settings", return_value=settings)


def mock_config(**kwargs):
    class MockConfig:
        mox_base = "unused"
        mora_base = "unused"
        client_id = "unused"
        client_secret = "unused"
        auth_server = "http://localhost:8081"
        auth_realm = "mo"

    instance = MockConfig()
    for name, val in kwargs.items():
        setattr(instance, name, val)

    return mock.patch.object(config, "get_config", return_value=instance)


def mock_create_mox_helper(module: ModuleType = initial):
    mock_mox_helper = mock.AsyncMock()
    mock_mox_helper.return_value = mock.AsyncMock()
    return mock.patch.object(module, "create_mox_helper", return_value=mock_mox_helper)



#!/usr/bin/env python3
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
from datetime import date
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

from pydantic import AnyHttpUrl
from pydantic import BaseSettings
from pydantic import conint
from pydantic import constr
from pydantic import Extra
from pydantic import Field
from pydantic import PositiveInt
from pydantic import SecretStr
from pydantic import UUID4
from ra_utils.load_settings import load_settings
from ra_utils.job_settings import JobSettings

from .models import JobFunction


class CommonSettings(BaseSettings):  # type: ignore
    """
    Settings common to both the SD importer and SD-changed-at
    """

    mora_base: AnyHttpUrl = Field("http://mo-service:5000")
    mox_base: AnyHttpUrl = Field("http://mo-service:5000/lora")

    sd_employment_field: Optional[str] = Field(default=None, regex="extension_[0-9]+")
    sd_global_from_date: date
    sd_import_run_db: str
    sd_import_too_deep: List[str] = []
    sd_institution_identifier: str
    sd_password: SecretStr
    sd_user: str
    sd_job_function: JobFunction
    sd_monthly_hourly_divide: PositiveInt

    # List of SD JobPositionIdentifiers which should not result in creation
    # of engagements
    sd_skip_employment_types: List[str] = []

    # If true, the <TelephoneNumberIdentifier> in <ContactInformation> in
    # <Employment> in the <Person> tag will be used to set a value that can
    # trigger AD-creation of the user
    # (see https://redmine.magenta-aps.dk/issues/56089)
    sd_phone_number_id_for_ad_creation: bool = False
    # SD can only set <TelephoneNumberIdentifier> to an integer value, hence
    # the value 14 is used (letters A=1 and D=4 for "AD")
    sd_phone_number_id_trigger: str = "14"
    sd_phone_number_id_for_ad_string: str = "AD-bruger fra SD"

    cpr_uuid_map_path: str = (
        "/opt/dipex/os2mo-data-import-and-export/settings/cpr_uuid_map.csv"
    )

    job_settings: JobSettings = JobSettings()


def gen_json_file_settings_func(settings_class: Type[CommonSettings]):
    def json_file_settings(settings: BaseSettings) -> Dict[str, Any]:  # type: ignore
        try:
            json_settings = load_settings()
        except FileNotFoundError:
            return dict()

        # Remove the "integrations.SD_Lon." part of the key name
        json_settings = {
            key.replace("integrations.SD_Lon.", "sd_"): value
            for key, value in json_settings.items()
        }

        # Remove any double "sd_sd_" in the keys
        json_settings = {
            key.replace("sd_sd_", "sd_"): value for key, value in json_settings.items()
        }

        # Replace dots with underscores to be Pydantic compliant
        json_settings = {
            key.replace(".", "_"): value for key, value in json_settings.items()
        }

        # Remove settings forbidden according to the Settings model
        properties = settings_class.schema()["properties"].keys()
        json_settings = {
            key: value for key, value in json_settings.items() if key in properties
        }

        return json_settings

    return json_file_settings


class SDCommonSettings(CommonSettings):
    class Config:
        extra = Extra.forbid

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                env_settings,
                gen_json_file_settings_func(SDCommonSettings),
                file_secret_settings,
            )


class ImporterSettings(CommonSettings):
    municipality_code: conint(ge=100, le=999)  # type: ignore
    municipality_name: str
    sd_importer_create_associations: bool = True
    sd_importer_create_email_addresses: bool = True

    # Deprecated
    sd_importer_employment_date_as_engagement_start_date: bool = False

    sd_no_salary_minimum_id: Optional[int] = None
    sd_use_ad_integration: bool = True

    class Config:
        extra = Extra.forbid

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                env_settings,
                gen_json_file_settings_func(ImporterSettings),
                file_secret_settings,
            )


class ChangedAtSettings(CommonSettings):
    sd_fix_departments_root: Optional[UUID4] = None
    sd_no_salary_minimum_id: Optional[int] = None
    sd_overwrite_existing_employment_name = True

    # List of CRPs to either include OR exclude in the run
    sd_cprs: List[constr(regex="^[0-9]{10}$")] = []  # type: ignore # noqa

    # If true, the sd_cprs will be excluded and if false, only the
    # CPRs in sd_cprs will be included in the run
    sd_exclude_cprs_mode: bool = True

    sd_read_forced_uuids: bool = True
    sd_update_primary_engagement: bool = True
    sd_use_ad_integration: bool = True

    class Config:
        extra = Extra.forbid

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                env_settings,
                gen_json_file_settings_func(ChangedAtSettings),
                file_secret_settings,
            )


@lru_cache()
def get_common_settings(*args, **kwargs) -> SDCommonSettings:
    return SDCommonSettings(*args, **kwargs)


@lru_cache()
def get_importer_settings(*args, **kwargs) -> ImporterSettings:
    return ImporterSettings(*args, **kwargs)


@lru_cache()
def get_changed_at_settings(*args, **kwargs) -> ChangedAtSettings:
    return ChangedAtSettings(*args, **kwargs)



# Denne snurrebasse har til hensigt at rette op på manglende navn-komponenter
# efter identificering af en fejl i sd changed at, som bevirkede at personer, som
# havde ændret navn (men kun fornavn ELLER efternavn vill stå med en blank
# værdi i den del af navnet, der IKKE var blevet opdateret.
#
# angiv uuider for de berørte på kommandolinien
import datetime
import sys

import click

from .config import get_changed_at_settings
from .sd_changed_at import ChangeAtSD
from .sd_common import sd_lookup


def read_person(cat, cpr):
    params = {
        "EffectiveDate": cat.from_date.strftime("%d.%m.%Y"),
        "PersonCivilRegistrationIdentifier": cpr,
        "StatusActiveIndicator": "true",
        "StatusPassiveIndicator": "true",
        "ContactInformationIndicator": "false",
        "PostalAddressIndicator": "false",
    }
    url = "GetPerson20111201"
    response = sd_lookup(url, params=params)
    person = response.get("Person", [])

    if not isinstance(person, list):
        person = [person]
    return person


@click.command()
@click.option("--uuid", type=click.UUID, required=True, multiple=True)
def cli(uuid):
    uuids = uuid
    settings = get_changed_at_settings()
    settings.sd_use_ad_integration = False
    cat = ChangeAtSD(settings, from_date=datetime.date.today())
    mh = cat.helper
    for uuid in uuids:
        mh = cat.helper
        mouser = mh.read_user(uuid)
        if mouser is None:
            raise ValueError
        sdnow = read_person(cat, mouser["cpr_no"])
        if len(sdnow) > 1:
            print(sdnow)
        else:
            person = sdnow[0]
            given_name = person.get("PersonGivenName", mouser.get("givenname", ""))
            sur_name = person.get("PersonSurnameName", mouser.get("surname", ""))
            payload = {
                "uuid": uuid,
                "givenname": given_name,
                "surname": sur_name,
                "cpr_no": mouser["cpr_no"],
                "org": {"uuid": mouser["org"]["uuid"]},
            }

            print(payload)
            mh._mo_post("e/create", payload).json()


if __name__ == "__main__":
    cli()



import typing
from collections import OrderedDict
from typing import Any
from typing import List
from typing import Tuple
from unittest.mock import MagicMock
from uuid import uuid4

import pytest

from ra_utils.attrdict import AttrDict
from ra_utils.attrdict import attrdict


@pytest.fixture
def mock_graphql_client():
    return MagicMock()


def get_sd_person_fixture(
    cpr: str, first_name: str, last_name: str, employment_id: str
) -> Tuple[AttrDict, List[typing.OrderedDict[str, Any]]]:
    """
    Get an SD person fixture. The function generates both the XML response
    returned from the SD API endpoint "GetPerson20111201" and the expected
    `OrderedDict` after parsing the XML.

    Args:
        cpr: The CPR number of the SD person.
        first_name: The first name (given name) of the SD person.
        last_name: The last name (surname) of the SD person.
        employment_id: The employmentID, e.g. 12345, of the SD person.

    Returns:
        Tuple with two elements. The first element is the raw XML response
        from the "GetPerson20111201" SD endpoint. The second element is the
        `OrderedDict` expected to be returned from get_sd_person.

    Example:
        ```
        >>> fix = get_sd_person_fixture('123456-1234', 'Bruce', 'Lee', "12345")
        >>> print(fix[0].text)
            <GetPerson20111201 creationDateTime="2020-12-03T17:40:10">
                <RequestStructure>
                    <InstitutionIdentifier>XX</InstitutionIdentifier>
                    <PersonCivilRegistrationIdentifier>123456-1234</PersonCivilRegistrationIdentifier>
                    <EffectiveDate>2020-12-03</EffectiveDate>
                    <StatusActiveIndicator>true</StatusActiveIndicator>
                    <StatusPassiveIndicator>false</StatusPassiveIndicator>
                    <ContactInformationIndicator>false</ContactInformationIndicator>
                    <PostalAddressIndicator>false</PostalAddressIndicator>
                </RequestStructure>
                <Person>
                    <PersonCivilRegistrationIdentifier>123456-1234</PersonCivilRegistrationIdentifier>
                    <PersonGivenName>Bruce</PersonGivenName>
                    <PersonSurnameName>Lee</PersonSurnameName>
                    <Employment>
                        <EmploymentIdentifier>12345</EmploymentIdentifier>
                    </Employment>
                </Person>
            </GetPerson20111201>
        >>> print(fix[1])
            [
                OrderedDict([
                    ('PersonCivilRegistrationIdentifier', '123456-1234'),
                    ('PersonGivenName', 'Bruce'),
                    ('PersonSurnameName', 'Lee'),
                    ('Employment', OrderedDict([
                        ('EmploymentIdentifier', '12345')
                    ]))
                ])
            ]
        ```
    """

    institution_id = "XX"

    sd_request_reply = attrdict(
        {
            "text": f"""
        <GetPerson20111201 creationDateTime="2020-12-03T17:40:10">
            <RequestStructure>
                <InstitutionIdentifier>{institution_id}</InstitutionIdentifier>
                <PersonCivilRegistrationIdentifier>{cpr}</PersonCivilRegistrationIdentifier>
                <EffectiveDate>2020-12-03</EffectiveDate>
                <StatusActiveIndicator>true</StatusActiveIndicator>
                <StatusPassiveIndicator>false</StatusPassiveIndicator>
                <ContactInformationIndicator>false</ContactInformationIndicator>
                <PostalAddressIndicator>false</PostalAddressIndicator>
            </RequestStructure>
            <Person>
                <PersonCivilRegistrationIdentifier>{cpr}</PersonCivilRegistrationIdentifier>
                <PersonGivenName>{first_name}</PersonGivenName>
                <PersonSurnameName>{last_name}</PersonSurnameName>
                <Employment>
                    <EmploymentIdentifier>{employment_id}</EmploymentIdentifier>
                </Employment>
            </Person>
        </GetPerson20111201>
        """
        }
    )

    expected_read_person_result = [
        OrderedDict(
            [
                ("PersonCivilRegistrationIdentifier", cpr),
                ("PersonGivenName", first_name),
                ("PersonSurnameName", last_name),
                (
                    "Employment",
                    OrderedDict([("EmploymentIdentifier", employment_id)]),
                ),
            ]
        )
    ]

    return sd_request_reply, expected_read_person_result


def read_employment_fixture(
    cpr: str, employment_id: str, job_id: str, job_title: str, status: str = "1"
) -> Tuple[AttrDict, List[typing.OrderedDict[str, Any]]]:
    """
    Get an SD employment fixture. The function is use for mocking calls to
    the "GetEmploymentChangedAtDate20111201" SD API endpoint, i.e. the endpoint
    that gets new changes *registered* between a from date and a to date. The
    function generates both the XML response endpoint
    "GetEmploymentChangedAtDate20111201" and the expected `OrderedDict` after
    parsing the XML.

    Args:
        cpr: The CPR number of the SD person.
        employment_id: The SD employment ID.
        job_id: The SD job ID.
        job_title: The SD profession.
        status: SD employment status.

    Returns:
        Tuple with two elements. The first element is the raw XML response
        from the "GetEmploymentChangedAtDate20111201" SD endpoint. The second
        element is the `OrderedDict` expected to be returned from
        read_employment_changed.

    Example:
        ```
        >>> fix=read_employment_fixture("123456-1234", "12345", "1", "chief", "1")
        >>> print(fix[1])
            [
                OrderedDict([
                    ('PersonCivilRegistrationIdentifier', '123456-1234'),
                    ('Employment', OrderedDict([
                        ('EmploymentIdentifier', '12345'),
                        ('EmploymentDate', '2020-11-10'),
                        ('AnniversaryDate', '2004-08-15'),
                        ('EmploymentDepartment', OrderedDict([
                            ('@changedAtDate', '2020-11-10'),
                            ('ActivationDate', '2020-11-10'),
                            ('DeactivationDate', '9999-12-31'),
                            ('DepartmentIdentifier', 'deprtment_id'),
                            ('DepartmentUUIDIdentifier', 'department_uuid')
                        ])),
                        ('Profession', OrderedDict([
                            ('@changedAtDate', '2020-11-10'),
                            ('ActivationDate', '2020-11-10'),
                            ('DeactivationDate', '9999-12-31'),
                            ('JobPositionIdentifier', '1'),
                            ('EmploymentName', 'chief'),
                            ('AppointmentCode', '0')
                        ])),
                        ('EmploymentStatus', [
                            OrderedDict([
                                ('@changedAtDate', '2020-11-10'),
                                ('ActivationDate', '2020-11-10'),
                                ('DeactivationDate', '2021-02-09'),
                                ('EmploymentStatusCode', '1')
                            ]),
                            OrderedDict([
                                ('@changedAtDate', '2020-11-10'),
                                ('ActivationDate', '2021-02-10'),
                                ('DeactivationDate', '9999-12-31'),
                                ('EmploymentStatusCode', '8')
                            ])
                        ])
                    ]))
                ])
            ]
        ```
    """
    institution_id = "institution_id"
    department_id = "deprtment_id"
    department_uuid = "department_uuid"

    sd_request_structure = f"""
        <RequestStructure>
            <InstitutionIdentifier>{institution_id}</InstitutionIdentifier>
            <ActivationDate>2020-11-01</ActivationDate>
            <ActivationTime>00:00:00</ActivationTime>
            <DeactivationDate>2020-12-02</DeactivationDate>
            <DeactivationTime>23:59:59</DeactivationTime>
            <DepartmentIndicator>true</DepartmentIndicator>
            <EmploymentStatusIndicator>true</EmploymentStatusIndicator>
            <ProfessionIndicator>true</ProfessionIndicator>
            <SalaryAgreementIndicator>false</SalaryAgreementIndicator>
            <SalaryCodeGroupIndicator>false</SalaryCodeGroupIndicator>
            <WorkingTimeIndicator>false</WorkingTimeIndicator>
            <UUIDIndicator>true</UUIDIndicator>
            <FutureInformationIndicator>false</FutureInformationIndicator>
        </RequestStructure>
    """
    sd_request_person_employeed = f"""
        <Person>
            <PersonCivilRegistrationIdentifier>{cpr}</PersonCivilRegistrationIdentifier>
            <Employment>
                <EmploymentIdentifier>{employment_id}</EmploymentIdentifier>
                <EmploymentDate>2020-11-10</EmploymentDate>
                <AnniversaryDate>2004-08-15</AnniversaryDate>
                <EmploymentDepartment changedAtDate="2020-11-10">
                    <ActivationDate>2020-11-10</ActivationDate>
                    <DeactivationDate>9999-12-31</DeactivationDate>
                    <DepartmentIdentifier>{department_id}</DepartmentIdentifier>
                    <DepartmentUUIDIdentifier>{department_uuid}</DepartmentUUIDIdentifier>
                </EmploymentDepartment>
                <Profession changedAtDate="2020-11-10">
                    <ActivationDate>2020-11-10</ActivationDate>
                    <DeactivationDate>9999-12-31</DeactivationDate>
                    <JobPositionIdentifier>{job_id}</JobPositionIdentifier>
                    <EmploymentName>{job_title}</EmploymentName>
                    <AppointmentCode>0</AppointmentCode>
                </Profession>
                <EmploymentStatus changedAtDate="2020-11-10">
                    <ActivationDate>2020-11-10</ActivationDate>
                    <DeactivationDate>2021-02-09</DeactivationDate>
                    <EmploymentStatusCode>1</EmploymentStatusCode>
                </EmploymentStatus>
                <EmploymentStatus changedAtDate="2020-11-10">
                    <ActivationDate>2021-02-10</ActivationDate>
                    <DeactivationDate>9999-12-31</DeactivationDate>
                    <EmploymentStatusCode>8</EmploymentStatusCode>
                </EmploymentStatus>
            </Employment>
        </Person>
    """
    employeed_result = OrderedDict(
        [
            ("PersonCivilRegistrationIdentifier", cpr),
            (
                "Employment",
                OrderedDict(
                    [
                        ("EmploymentIdentifier", employment_id),
                        ("EmploymentDate", "2020-11-10"),
                        ("AnniversaryDate", "2004-08-15"),
                        (
                            "EmploymentDepartment",
                            OrderedDict(
                                [
                                    ("@changedAtDate", "2020-11-10"),
                                    ("ActivationDate", "2020-11-10"),
                                    ("DeactivationDate", "9999-12-31"),
                                    ("DepartmentIdentifier", department_id),
                                    (
                                        "DepartmentUUIDIdentifier",
                                        department_uuid,
                                    ),
                                ]
                            ),
                        ),
                        (
                            "Profession",
                            OrderedDict(
                                [
                                    ("@changedAtDate", "2020-11-10"),
                                    ("ActivationDate", "2020-11-10"),
                                    ("DeactivationDate", "9999-12-31"),
                                    ("JobPositionIdentifier", job_id),
                                    ("EmploymentName", job_title),
                                    ("AppointmentCode", "0"),
                                ]
                            ),
                        ),
                        (
                            "EmploymentStatus",
                            [
                                OrderedDict(
                                    [
                                        ("@changedAtDate", "2020-11-10"),
                                        ("ActivationDate", "2020-11-10"),
                                        ("DeactivationDate", "2021-02-09"),
                                        ("EmploymentStatusCode", "1"),
                                    ]
                                ),
                                OrderedDict(
                                    [
                                        ("@changedAtDate", "2020-11-10"),
                                        ("ActivationDate", "2021-02-10"),
                                        ("DeactivationDate", "9999-12-31"),
                                        ("EmploymentStatusCode", "8"),
                                    ]
                                ),
                            ],
                        ),
                    ]
                ),
            ),
        ]
    )
    sd_request_person_deleted = f"""
        <Person>
            <PersonCivilRegistrationIdentifier>{cpr}</PersonCivilRegistrationIdentifier>
            <Employment>
                <EmploymentIdentifier>{employment_id}</EmploymentIdentifier>
                <EmploymentStatus changedAtDate="2020-11-09">
                    <ActivationDate>2020-11-01</ActivationDate>
                    <DeactivationDate>9999-12-31</DeactivationDate>
                    <EmploymentStatusCode>S</EmploymentStatusCode>
                </EmploymentStatus>
            </Employment>
        </Person>
    """
    deleted_result = OrderedDict(
        [
            ("PersonCivilRegistrationIdentifier", cpr),
            (
                "Employment",
                OrderedDict(
                    [
                        ("EmploymentIdentifier", employment_id),
                        (
                            "EmploymentStatus",
                            OrderedDict(
                                [
                                    ("@changedAtDate", "2020-11-09"),
                                    ("ActivationDate", "2020-11-01"),
                                    ("DeactivationDate", "9999-12-31"),
                                    ("EmploymentStatusCode", "S"),
                                ]
                            ),
                        ),
                    ]
                ),
            ),
        ]
    )

    person_table = {
        "1": (sd_request_person_employeed, employeed_result),
        "S": (sd_request_person_deleted, deleted_result),
    }
    sd_response = (
        """
        <GetEmploymentChangedAtDate20111201 creationDateTime="2020-12-02T16:44:19">
        """
        + sd_request_structure
        + person_table[status][0]
        + """
        </GetEmploymentChangedAtDate20111201>
    """
    )
    sd_request_reply = attrdict({"text": sd_response})
    expected_read_employment_result = [person_table[status][1]]
    return sd_request_reply, expected_read_employment_result


def get_department_fixture(
    institution_id="XX",
    institution_uuid=None,
    region_id="XY",
    region_uuid=None,
    department1_id="D1X",
    department1_uuid=None,
    department1_name="D1X-name",
    sub_department1_id="D1Y",
    sub_department1_uuid=None,
    sub_department1_name="D1Y-name",
    department2_id="D2X",
    department2_uuid=None,
    department2_name="D2X-name",
    sub_department2_id="D2Y",
    sub_department2_uuid=None,
    sub_department2_name="D2Y-name",
):
    institution_uuid = institution_uuid or str(uuid4())
    region_uuid = region_uuid or str(uuid4())

    department1_uuid = department1_uuid or str(uuid4())
    sub_department1_uuid = sub_department1_uuid or str(uuid4())

    department2_uuid = department2_uuid or str(uuid4())
    sub_department2_uuid = sub_department2_uuid or str(uuid4())

    return attrdict(
        {
            "text": f"""
        <GetDepartment20111201 creationDateTime="2021-10-20T14:51:23">
        <RequestStructure>
            <InstitutionIdentifier>{institution_id}</InstitutionIdentifier>
            <ActivationDate>2021-10-20</ActivationDate>
            <DeactivationDate>2021-10-20</DeactivationDate>
            <ContactInformationIndicator>true</ContactInformationIndicator>
            <DepartmentNameIndicator>true</DepartmentNameIndicator>
            <EmploymentDepartmentIndicator>true</EmploymentDepartmentIndicator>
            <PostalAddressIndicator>true</PostalAddressIndicator>
            <ProductionUnitIndicator>true</ProductionUnitIndicator>
            <UUIDIndicator>true</UUIDIndicator>
        </RequestStructure>
        <RegionIdentifier>{region_id}</RegionIdentifier>
        <RegionUUIDIdentifier>{region_uuid}</RegionUUIDIdentifier>
        <InstitutionIdentifier>{institution_id}</InstitutionIdentifier>
        <InstitutionUUIDIdentifier>{institution_uuid}</InstitutionUUIDIdentifier>
        <Department>
            <ActivationDate>2010-01-01</ActivationDate>
            <DeactivationDate>9999-12-31</DeactivationDate>
            <DepartmentIdentifier>{department1_id}</DepartmentIdentifier>
            <DepartmentUUIDIdentifier>{department1_uuid}</DepartmentUUIDIdentifier>
            <DepartmentLevelIdentifier>Afdelings-niveau</DepartmentLevelIdentifier>
            <DepartmentName>{department1_name}</DepartmentName>
            <PostalAddress>
                <StandardAddressIdentifier>
                    Department 1 Address
                </StandardAddressIdentifier>
                <PostalCode>8600</PostalCode>
                <DistrictName>Silkeborg</DistrictName>
                <MunicipalityCode>0740</MunicipalityCode>
            </PostalAddress>
        </Department>
        <Department>
            <ActivationDate>2011-01-01</ActivationDate>
            <DeactivationDate>9999-12-31</DeactivationDate>
            <DepartmentIdentifier>{sub_department1_id}</DepartmentIdentifier>
            <DepartmentUUIDIdentifier>{sub_department1_uuid}</DepartmentUUIDIdentifier>
            <DepartmentLevelIdentifier>NY5-niveau</DepartmentLevelIdentifier>
            <DepartmentName>{sub_department1_name}</DepartmentName>
            <PostalAddress>
                <StandardAddressIdentifier>
                    Sub Department 1 Address
                </StandardAddressIdentifier>
                <PostalCode>8600</PostalCode>
                <DistrictName>Silkeborg</DistrictName>
                <MunicipalityCode>0740</MunicipalityCode>
            </PostalAddress>
            <ContactInformation>
                <EmailAddressIdentifier>sub_department_1@example.org</EmailAddressIdentifier>
                <EmailAddressIdentifier>Empty@Empty</EmailAddressIdentifier>
            </ContactInformation>
        </Department>
        <Department>
            <ActivationDate>2012-01-01</ActivationDate>
            <DeactivationDate>9999-12-31</DeactivationDate>
            <DepartmentIdentifier>{department2_id}</DepartmentIdentifier>
            <DepartmentUUIDIdentifier>{department2_uuid}</DepartmentUUIDIdentifier>
            <DepartmentLevelIdentifier>Afdelings-niveau</DepartmentLevelIdentifier>
            <DepartmentName>{department2_name}</DepartmentName>
            <PostalAddress>
                <StandardAddressIdentifier>
                    Department 2 Address
                </StandardAddressIdentifier>
                <PostalCode>8600</PostalCode>
                <DistrictName>Silkeborg</DistrictName>
                <MunicipalityCode>0740</MunicipalityCode>
            </PostalAddress>
        </Department>
        <Department>
            <ActivationDate>2013-01-01</ActivationDate>
            <DeactivationDate>9999-12-31</DeactivationDate>
            <DepartmentIdentifier>{sub_department2_id}</DepartmentIdentifier>
            <DepartmentUUIDIdentifier>{sub_department2_uuid}</DepartmentUUIDIdentifier>
            <DepartmentLevelIdentifier>NY5-niveau</DepartmentLevelIdentifier>
            <DepartmentName>{sub_department2_name}</DepartmentName>
            <PostalAddress>
                <StandardAddressIdentifier>
                    Sub Department 2 Address
                </StandardAddressIdentifier>
                <PostalCode>8600</PostalCode>
                <DistrictName>Silkeborg</DistrictName>
                <MunicipalityCode>0740</MunicipalityCode>
            </PostalAddress>
            <ContactInformation>
                <EmailAddressIdentifier>sub_department_2@example.org</EmailAddressIdentifier>
                <EmailAddressIdentifier>Empty@Empty</EmailAddressIdentifier>
            </ContactInformation>
        </Department>
        </GetDepartment20111201>
        """
        }
    )


def get_organisation_fixture(
    institution_id="XX",
    institution_uuid=None,
    region_id="XY",
    region_uuid=None,
    department_structure_name="XX-Basis",
    department1_id="D1X",
    department1_uuid=None,
    sub_department1_id="D1Y",
    sub_department1_uuid=None,
    department2_id="D2X",
    department2_uuid=None,
    sub_department2_id="D2Y",
    sub_department2_uuid=None,
):
    institution_uuid = institution_uuid or str(uuid4())
    region_uuid = region_uuid or str(uuid4())

    department1_uuid = department1_uuid or str(uuid4())
    sub_department1_uuid = sub_department1_uuid or str(uuid4())

    department2_uuid = department2_uuid or str(uuid4())
    sub_department2_uuid = sub_department2_uuid or str(uuid4())

    return attrdict(
        {
            "text": f"""
        <GetOrganization20111201 creationDateTime="2021-10-20T14:35:44">
            <RequestStructure>
                <InstitutionIdentifier>{institution_id}</InstitutionIdentifier>
                <ActivationDate>2021-10-20</ActivationDate>
                <DeactivationDate>2021-10-20</DeactivationDate>
                <UUIDIndicator>true</UUIDIndicator>
            </RequestStructure>
            <RegionIdentifier>{region_id}</RegionIdentifier>
            <RegionUUIDIdentifier>{region_uuid}</RegionUUIDIdentifier>
            <InstitutionIdentifier>{institution_id}</InstitutionIdentifier>
            <InstitutionUUIDIdentifier>{institution_uuid}</InstitutionUUIDIdentifier>
            <DepartmentStructureName>{department_structure_name}</DepartmentStructureName>
            <OrganizationStructure>
                <DepartmentLevelReference>
                    <DepartmentLevelIdentifier>Afdelings-niveau</DepartmentLevelIdentifier>
                    <DepartmentLevelReference>
                        <DepartmentLevelIdentifier>NY0-niveau</DepartmentLevelIdentifier>
                        <DepartmentLevelReference>
                            <DepartmentLevelIdentifier>NY1-niveau</DepartmentLevelIdentifier>
                            <DepartmentLevelReference>
                                <DepartmentLevelIdentifier>NY2-niveau</DepartmentLevelIdentifier>
                                <DepartmentLevelReference>
                                    <DepartmentLevelIdentifier>NY3-niveau</DepartmentLevelIdentifier>
                                    <DepartmentLevelReference>
                                        <DepartmentLevelIdentifier>NY4-niveau</DepartmentLevelIdentifier>
                                        <DepartmentLevelReference>
                                            <DepartmentLevelIdentifier>NY5-niveau</DepartmentLevelIdentifier>
                                            <DepartmentLevelReference>
                                                <DepartmentLevelIdentifier>NY6-niveau</DepartmentLevelIdentifier>
                                            </DepartmentLevelReference>
                                        </DepartmentLevelReference>
                                    </DepartmentLevelReference>
                                </DepartmentLevelReference>
                            </DepartmentLevelReference>
                        </DepartmentLevelReference>
                    </DepartmentLevelReference>
                </DepartmentLevelReference>
            </OrganizationStructure>
            <Organization>
                <ActivationDate>2021-10-01</ActivationDate>
                <DeactivationDate>2021-10-20</DeactivationDate>
                <DepartmentReference>
                    <DepartmentIdentifier>{department1_id}</DepartmentIdentifier>
                    <DepartmentUUIDIdentifier>{department1_uuid}</DepartmentUUIDIdentifier>
                    <DepartmentLevelIdentifier>Afdelings-niveau</DepartmentLevelIdentifier>
                    <DepartmentReference>
                        <DepartmentIdentifier>{sub_department1_id}</DepartmentIdentifier>
                        <DepartmentUUIDIdentifier>{sub_department1_uuid}</DepartmentUUIDIdentifier>
                        <DepartmentLevelIdentifier>NY5-niveau</DepartmentLevelIdentifier>
                    </DepartmentReference>
                </DepartmentReference>
                <DepartmentReference>
                    <DepartmentIdentifier>{department2_id}</DepartmentIdentifier>
                    <DepartmentUUIDIdentifier>{department2_uuid}</DepartmentUUIDIdentifier>
                    <DepartmentLevelIdentifier>Afdelings-niveau</DepartmentLevelIdentifier>
                    <DepartmentReference>
                        <DepartmentIdentifier>{sub_department2_id}</DepartmentIdentifier>
                        <DepartmentUUIDIdentifier>{sub_department2_uuid}</DepartmentUUIDIdentifier>
                        <DepartmentLevelIdentifier>NY5-niveau</DepartmentLevelIdentifier>
                    </DepartmentReference>
                </DepartmentReference>
            </Organization>
        </GetOrganization20111201>
        """
        }
    )


def get_employment_fixture(
    cpr, employment_id, department_id, department_uuid, job_pos_id, job_title
):
    """
    Get the OrderedDict returned when sd_common.sd_lookup is called for the
    SD endpoint GetEmployment20111201.
    """

    return OrderedDict(
        [
            (
                "Person",
                OrderedDict(
                    [
                        ("PersonCivilRegistrationIdentifier", cpr),
                        (
                            "Employment",
                            OrderedDict(
                                [
                                    ("EmploymentIdentifier", employment_id),
                                    ("EmploymentDate", "2020-11-10"),
                                    ("AnniversaryDate", "2004-08-15"),
                                    (
                                        "EmploymentDepartment",
                                        OrderedDict(
                                            [
                                                ("@changedAtDate", "2020-11-10"),
                                                ("ActivationDate", "2020-11-10"),
                                                ("DeactivationDate", "9999-12-31"),
                                                ("DepartmentIdentifier", department_id),
                                                (
                                                    "DepartmentUUIDIdentifier",
                                                    department_uuid,
                                                ),
                                            ]
                                        ),
                                    ),
                                    (
                                        "Profession",
                                        OrderedDict(
                                            [
                                                ("@changedAtDate", "2020-11-10"),
                                                ("ActivationDate", "2020-11-10"),
                                                ("DeactivationDate", "9999-12-31"),
                                                ("JobPositionIdentifier", job_pos_id),
                                                ("EmploymentName", job_title),
                                                ("AppointmentCode", "0"),
                                            ]
                                        ),
                                    ),
                                    (
                                        "EmploymentStatus",
                                        OrderedDict(
                                            [
                                                ("@changedAtDate", "2020-11-10"),
                                                ("ActivationDate", "2020-11-10"),
                                                ("DeactivationDate", "2021-02-09"),
                                                ("EmploymentStatusCode", "1"),
                                            ]
                                        ),
                                    ),
                                ]
                            ),
                        ),
                    ]
                ),
            )
        ]
    )


def get_read_employment_changed_fixture(
    cpr: int = 1234561234,
    employment_id: str = "12345",
    employment_date: str = "2020s-01-01",
    anniversary_date: str = "2020-01-01",
    employment_department_activation_date: str = "2020-01-01",
    employment_department_deactivation_date: str = "9999-12-31",
    department_id: str = "department_id",
    department_uuid_id: str = "department_uuid",
    profession_activation_date: str = "2020-01-01",
    profession_deactivation_date: str = "9999-12-31",
    job_pos_id: int = 1000,
    employment_name: str = "Employment name",
    appointment_code: str = "0",
    employment_status_activation_date: str = "2020-01-01",
    employment_status_deactivation_date: str = "9999-12-31",
    employment_status_code: int = 1,
):
    return [
        OrderedDict(
            [
                ("PersonCivilRegistrationIdentifier", str(cpr)),
                (
                    "Employment",
                    OrderedDict(
                        [
                            ("EmploymentIdentifier", employment_id),
                            ("EmploymentDate", employment_date),
                            ("AnniversaryDate", anniversary_date),
                            (
                                "EmploymentDepartment",
                                OrderedDict(
                                    [
                                        ("@changedAtDate", "2020-11-10"),
                                        (
                                            "ActivationDate",
                                            employment_department_activation_date,
                                        ),
                                        (
                                            "DeactivationDate",
                                            employment_department_deactivation_date,
                                        ),
                                        ("DepartmentIdentifier", department_id),
                                        (
                                            "DepartmentUUIDIdentifier",
                                            department_uuid_id,
                                        ),
                                    ]
                                ),
                            ),
                            (
                                "Profession",
                                OrderedDict(
                                    [
                                        ("@changedAtDate", "2020-11-10"),
                                        ("ActivationDate", profession_activation_date),
                                        (
                                            "DeactivationDate",
                                            profession_deactivation_date,
                                        ),
                                        ("JobPositionIdentifier", str(job_pos_id)),
                                        ("EmploymentName", employment_name),
                                        ("AppointmentCode", appointment_code),
                                    ]
                                ),
                            ),
                            (
                                "EmploymentStatus",
                                OrderedDict(
                                    [
                                        ("@changedAtDate", "2020-11-10"),
                                        (
                                            "ActivationDate",
                                            employment_status_activation_date,
                                        ),
                                        (
                                            "DeactivationDate",
                                            employment_status_deactivation_date,
                                        ),
                                        (
                                            "EmploymentStatusCode",
                                            str(employment_status_code),
                                        ),
                                    ]
                                ),
                            ),
                        ]
                    ),
                ),
            ]
        )
    ]



from typing import Any
from typing import Dict
from typing import Optional
from unittest.mock import MagicMock
from unittest.mock import patch
from uuid import UUID
from uuid import uuid4

from anytree import Node
import hypothesis.strategies as st
import pytest
from hypothesis import given
from hypothesis import settings
from os2mo_data_import import ImportHelper
from os2mo_data_import.mora_data_types import OrganisationUnitType
from parameterized import parameterized
from ra_utils.attrdict import attrdict

from .fixtures import get_department_fixture
from .fixtures import get_organisation_fixture
from sdlon.config import ImporterSettings
from sdlon.sd_importer import SdImport


@pytest.fixture(autouse=True)
def mock_json(monkeypatch):
    monkeypatch.setattr("sdlon.config.load_settings", lambda: dict())


class SdImportTest(SdImport):
    def __init__(self, *args, **kwargs):
        self.add_people_mock = MagicMock()
        self.read_department_info_mock = MagicMock()

        self.orig_add_people = self.add_people
        self.orig_read_department_info = self._read_department_info

        self.add_people = self.add_people_mock
        self._read_department_info = self.read_department_info_mock
        self._get_institution = MagicMock(
            return_value=UUID("00000000-0000-0000-0000-000000000000")
        )

        super().__init__(*args, **kwargs)


def get_sd_importer(
    municipality_name: str = "Andeby Kommune",
    municipality_code: int = 100,
    org_only: bool = False,
    override_settings: Optional[Dict[str, Any]] = None,
) -> SdImportTest:
    override_settings = override_settings or {}

    mora_base = "http://mora/"
    mox_base = "http://mox/"

    importer = ImportHelper(
        create_defaults=True,
        mox_base=mox_base,
        mora_base=mora_base,
        seperate_names=True,
    )

    settings_dict = {
        "municipality_name": municipality_name,
        "municipality_code": municipality_code,
        "sd_global_from_date": "1970-01-01",
        "sd_employment_field": "extension_1",
        "sd_import_run_db": "run_db.sqlite",
        "sd_institution_identifier": "XY",
        "sd_job_function": "EmploymentName",
        "sd_monthly_hourly_divide": 9000,
        "sd_password": "secret",
        "sd_user": "user",
        "sd_importer_create_associations": False,
    }
    settings_dict.update(override_settings)
    settings = ImporterSettings.parse_obj(settings_dict)

    sd = SdImportTest(importer, settings=settings, org_only=org_only, ad_info=None)

    # add_people should not be called when rg-only
    if org_only:
        sd.add_people_mock.assert_not_called()
    else:
        sd.add_people_mock.assert_called_once()
    sd.read_department_info_mock.assert_called_once()

    assert sd.importer.organisation is not None
    assert sd.importer.organisation[0] == municipality_name
    assert sd.importer.organisation[1].uuid == "00000000-0000-0000-0000-000000000000"
    assert sd.importer.organisation[1].name == municipality_name
    assert sd.importer.organisation[1].user_key == municipality_name
    assert sd.importer.organisation[1].municipality_code == str(municipality_code)
    assert sd.importer.organisation[1].date_from == "1930-01-01"
    assert sd.importer.organisation[1].date_to == "infinity"

    assert sd.importer.klassifikation is not None
    assert sd.importer.klassifikation[0] == municipality_name
    assert sd.importer.klassifikation[1].organisation_uuid is None
    assert sd.importer.klassifikation[1].parent_name == municipality_name
    assert sd.importer.klassifikation[1].user_key == municipality_name
    assert sd.importer.klassifikation[1].description == "umbrella"
    assert sd.importer.klassifikation[1].date_from == "1930-01-01"
    assert sd.importer.klassifikation[1].date_to == "infinity"

    # 29 classes exist hardcoded within sd_importer.py
    assert len(sd.importer.klasse_objects) == 30

    # 18 facets in os2mo_data_import/defaults.py
    assert len(sd.importer.facet_objects) == 18

    # No other objects
    assert len(sd.importer.addresses) == 0
    assert len(sd.importer.itsystems) == 0
    assert len(sd.importer.organisation_units) == 0
    assert len(sd.importer.organisation_unit_details) == 0
    assert len(sd.importer.employees) == 0
    assert len(sd.importer.employee_details) == 0

    return sd


@given(st.text(), st.integers(min_value=100, max_value=999), st.booleans())
def test_instantiation(municipality_name: str, municipality_code: int, org_only: bool):
    get_sd_importer(municipality_name, municipality_code, org_only)


@given(st.booleans())
def test_create_employee(create_associations: bool):
    sd = get_sd_importer(
        override_settings={
            "sd_importer_create_associations": create_associations,
        }
    )

    org_unit_uuid = str(uuid4())

    sd.nodes[org_unit_uuid] = attrdict({"name": "org_unit", "date_from": "1900-01-01"})
    sd.importer.organisation_units[org_unit_uuid] = OrganisationUnitType(
        name="org_unit",
        type_ref="type_ref",
        date_from="1900-01-01",
    )

    original_classes = set(sd.importer.klasse_objects.keys())

    cpr_no = "0101709999"
    sd.importer.add_employee(
        name=("given_name", "sur_name"),
        identifier=cpr_no,
        cpr_no=cpr_no,
        user_key="employee_user_key",
        uuid="employee_uuid",
    )
    sd.create_employee(
        {
            "PersonCivilRegistrationIdentifier": cpr_no,
            "Employment": [
                {
                    "Profession": {"JobPositionIdentifier": "job_id_123"},
                    "AnniversaryDate": "2004-08-15",
                    "EmploymentStatus": {
                        "EmploymentStatusCode": "1",
                        "ActivationDate": "1970-01-01",
                        "DeactivationDate": "9999-12-31",
                    },
                    "EmploymentIdentifier": "TEST123",
                    "WorkingTime": {"OccupationRate": 1},
                    "EmploymentDepartment": {
                        "DepartmentUUIDIdentifier": org_unit_uuid,
                    },
                }
            ],
        }
    )

    new_classes = {
        key: value
        for key, value in sd.importer.klasse_objects.items()
        if key not in original_classes
    }
    assert len(new_classes) == 2
    engagement_type = new_classes["engagement_typejob_id_123"]
    job_id = new_classes["job_id_123"]
    assert engagement_type.date_from == job_id.date_from == "1930-01-01"
    assert engagement_type.date_to == job_id.date_to == "infinity"
    assert engagement_type.description == job_id.description is None
    assert engagement_type.scope == job_id.scope == "TEXT"
    assert engagement_type.example == job_id.example is None
    assert engagement_type.organisation_uuid == job_id.organisation_uuid is None
    assert engagement_type.facet_uuid == job_id.facet_uuid is None

    assert engagement_type.user_key == "engagement_typejob_id_123"
    assert engagement_type.title == "job_id_123"
    assert engagement_type.facet_type_ref == "engagement_type"
    UUID(engagement_type.uuid)

    assert job_id.user_key == "job_id_123"
    assert job_id.title == "job_id_123"
    assert job_id.facet_type_ref == "engagement_job_function"
    UUID(job_id.uuid)

    # 18 facets in os2mo_data_import/defaults.py
    assert len(sd.importer.facet_objects) == 18

    # None of these objects
    assert len(sd.importer.addresses) == 0
    assert len(sd.importer.itsystems) == 0
    assert len(sd.importer.organisation_unit_details) == 0

    # But one of these
    assert len(sd.importer.organisation_units) == 1
    assert len(sd.importer.employees) == 1
    employee = sd.importer.employees[cpr_no]
    assert employee.givenname == "given_name"
    assert employee.surname == "sur_name"
    assert employee.cpr_no == cpr_no
    assert employee.user_key == "employee_user_key"

    assert len(sd.importer.employee_details) == 1
    details = sd.importer.employee_details[cpr_no]

    if create_associations:
        # We expect one engagement, and one association
        assert len(details) == 2
        association, engagement = details

        assert association.type_id == "association"
        assert association.date_from == "1970-01-01"
        assert association.date_to is None
        assert association.user_key == "TEST123"
        assert association.org_unit_ref == org_unit_uuid
        assert association.type_ref == "SD-medarbejder"
    else:
        # We expect just an engagement
        assert len(details) == 1
        engagement = details[0]

    assert engagement.type_id == "engagement"
    assert engagement.date_from == "1970-01-01"
    assert engagement.date_to is None
    assert engagement.user_key == "TEST123"
    assert engagement.fraction == 1000000
    assert engagement.primary_ref == "non-primary"
    assert engagement.org_unit_ref == org_unit_uuid
    assert engagement.type_ref == "engagement_typejob_id_123"
    assert engagement.job_function_ref == "job_id_123"


@patch("sdlon.sd_common.sd_lookup_settings")
@patch("sdlon.sd_common.requests.get")
@patch("integrations.dawa_helper.dawa_lookup")
@given(st.booleans())
@settings(deadline=None)
def test_create_ou_tree(
    dawa_lookup, requests_get, sd_settings, create_email_addresses: bool
):
    sd_settings.return_value = ("", "", "")
    dawa_lookup.return_value = None

    sd = get_sd_importer(
        override_settings={
            "sd_importer_create_email_addresses": create_email_addresses,
        }
    )
    institution_uuid = str(uuid4())
    region_uuid = str(uuid4())
    department1_uuid = str(uuid4())
    sub_department1_uuid = str(uuid4())
    department2_uuid = str(uuid4())
    sub_department2_uuid = str(uuid4())

    original_classes = set(sd.importer.klasse_objects.keys())

    requests_get.return_value = get_department_fixture(
        institution_uuid=institution_uuid,
        region_uuid=region_uuid,
        department1_uuid=department1_uuid,
        sub_department1_uuid=sub_department1_uuid,
        department2_uuid=department2_uuid,
        sub_department2_uuid=sub_department2_uuid,
    )
    sd.info = sd.orig_read_department_info()

    org_unit_level_classes = {
        key: value
        for key, value in sd.importer.klasse_objects.items()
        if key not in original_classes
    }
    assert len(org_unit_level_classes) == 2
    afdelings_niveau = org_unit_level_classes["Afdelings-niveau"]
    ny5_niveau = org_unit_level_classes["NY5-niveau"]

    assert afdelings_niveau.date_from == ny5_niveau.date_from == "1930-01-01"
    assert afdelings_niveau.date_to == ny5_niveau.date_to == "infinity"
    assert afdelings_niveau.description == ny5_niveau.description is None
    assert afdelings_niveau.scope == ny5_niveau.scope == "TEXT"
    assert afdelings_niveau.example == ny5_niveau.example is None
    assert afdelings_niveau.organisation_uuid == ny5_niveau.organisation_uuid is None
    assert afdelings_niveau.facet_uuid == ny5_niveau.facet_uuid is None
    assert (
        afdelings_niveau.facet_type_ref == ny5_niveau.facet_type_ref == "org_unit_level"
    )

    assert afdelings_niveau.user_key == "Afdelings-niveau"
    assert afdelings_niveau.title == "Afdelings-niveau"
    UUID(afdelings_niveau.uuid)

    assert ny5_niveau.user_key == "NY5-niveau"
    assert ny5_niveau.title == "NY5-niveau"
    UUID(ny5_niveau.uuid)

    requests_get.return_value = get_organisation_fixture(
        institution_uuid=institution_uuid,
        region_uuid=region_uuid,
        department1_uuid=department1_uuid,
        sub_department1_uuid=sub_department1_uuid,
        department2_uuid=department2_uuid,
        sub_department2_uuid=sub_department2_uuid,
    )
    sd.create_ou_tree(create_orphan_container=False, sub_tree=None, super_unit=None)

    assert sd.importer.organisation is not None
    assert sd.importer.klassifikation is not None
    assert len(sd.importer.facet_objects) == 18

    # None of these objects
    assert len(sd.importer.addresses) == 0
    assert len(sd.importer.itsystems) == 0
    assert len(sd.importer.employees) == 0
    assert len(sd.importer.employee_details) == 0

    # But some of these
    assert len(sd.importer.organisation_units) == 5
    department1 = sd.importer.organisation_units[department1_uuid]
    sub_department1 = sd.importer.organisation_units[sub_department1_uuid]
    department2 = sd.importer.organisation_units[department2_uuid]
    sub_department2 = sd.importer.organisation_units[sub_department2_uuid]

    assert department1.date_from == "2010-01-01"
    assert department1.date_to is None
    assert department1.type_ref == "Enhed"

    assert sub_department1.date_from == "2011-01-01"
    assert sub_department1.date_to is None
    assert sub_department1.type_ref == "Enhed"

    assert department2.date_from == "2012-01-01"
    assert department2.date_to is None
    assert department2.type_ref == "Enhed"

    assert sub_department2.date_from == "2013-01-01"
    assert sub_department2.date_to is None
    assert sub_department2.type_ref == "Enhed"

    assert department1.name == "D1X-name"
    assert department1.user_key == "D1X"
    assert department1.org_unit_level_ref == "Afdelings-niveau"
    UUID(department1.uuid)

    assert sub_department1.name == "D1Y-name"
    assert sub_department1.user_key == "D1Y"
    assert sub_department1.org_unit_level_ref == "NY5-niveau"
    UUID(sub_department1.uuid)

    assert department2.name == "D2X-name"
    assert department2.user_key == "D2X"
    assert department2.org_unit_level_ref == "Afdelings-niveau"
    UUID(department2.uuid)

    assert sub_department2.name == "D2Y-name"
    assert sub_department2.user_key == "D2Y"
    assert sub_department2.org_unit_level_ref == "NY5-niveau"
    UUID(sub_department2.uuid)

    assert len(sd.importer.organisation_unit_details) == 5
    details = list(sd.importer.organisation_unit_details.values())
    (
        historic_details,
        department1_details,
        sub_department1_details,
        department2_details,
        sub_department2_details,
    ) = details

    assert department1_details == []
    assert department2_details == []

    if not create_email_addresses:
        assert sub_department1_details == []
        assert sub_department2_details == []
    else:
        assert len(sub_department1_details) == 1
        sub_department1_address = sub_department1_details[0]
        assert sub_department1_address.type_id == "address"
        assert sub_department1_address.date_from == "2011-01-01"
        assert sub_department1_address.date_to is None
        assert sub_department1_address.value == "sub_department_1@example.org"
        assert sub_department1_address.type_ref == "EmailUnit"

        assert len(sub_department2_details) == 1
        sub_department2_address = sub_department2_details[0]
        assert sub_department2_address.type_id == "address"
        assert sub_department2_address.date_from == "2013-01-01"
        assert sub_department2_address.date_to is None
        assert sub_department2_address.value == "sub_department_2@example.org"
        assert sub_department2_address.type_ref == "EmailUnit"


@patch("sdlon.sd_importer.uuid.uuid4")
def test_set_engagement_on_leave(mock_uuid4):

    # Arrange

    mock_uuid4.return_value = UUID("00000000-0000-0000-0000-000000000000")
    sd = get_sd_importer()
    org_unit_uuid = str(uuid4())

    sd.nodes[org_unit_uuid] = attrdict({"name": "org_unit", "date_from": "1900-01-01"})
    sd.importer.organisation_units[org_unit_uuid] = OrganisationUnitType(
        name="org_unit",
        type_ref="type_ref",
        date_from="1900-01-01",
    )

    cpr_no = "0101709999"
    sd.importer.add_employee(
        name=("given_name", "sur_name"),
        identifier=cpr_no,
        cpr_no=cpr_no,
        user_key="employee_user_key",
        uuid="employee_uuid",
    )

    # Act

    # Create an employee on leave (SD EmploymentStatusCode = 3)
    sd.create_employee(
        {
            "PersonCivilRegistrationIdentifier": cpr_no,
            "Employment": [
                {
                    "EmploymentDate": "1960-01-01",
                    "AnniversaryDate": "2004-08-15",
                    "Profession": {"JobPositionIdentifier": "job_id_123"},
                    "EmploymentStatus": {
                        "EmploymentStatusCode": "3",
                        "ActivationDate": "1970-01-01",
                        "DeactivationDate": "9999-12-31",
                    },
                    "EmploymentIdentifier": "TEST123",
                    "WorkingTime": {"OccupationRate": 1},
                    "EmploymentDepartment": {
                        "DepartmentUUIDIdentifier": org_unit_uuid,
                    },
                }
            ],
        }
    )

    # Assert

    details = sd.importer.employee_details[cpr_no]
    engagement, leave = details

    assert len(details) == 2
    assert engagement.type_id == "engagement"
    assert leave.type_id == "leave"

    assert engagement.uuid == "00000000-0000-0000-0000-000000000000"
    assert leave.engagement_uuid == "00000000-0000-0000-0000-000000000000"
    assert leave.date_from == "1970-01-01"


@patch("sdlon.sd_importer.uuid.uuid4")
def test_manager_dates_set_correctly(mock_uuid4):
    # Arrange

    mock_uuid4.return_value = UUID("00000000-0000-0000-0000-000000000000")
    sd = get_sd_importer()
    org_unit_uuid = str(uuid4())

    ou = attrdict({"name": "org_unit", "date_from": "1960-01-01"})

    sd.nodes[org_unit_uuid] = ou
    sd.importer.organisation_units[org_unit_uuid] = ou

    cpr_no = "0101709999"
    sd.importer.add_employee(
        name=("given_name", "sur_name"),
        identifier=cpr_no,
        cpr_no=cpr_no,
        user_key="employee_user_key",
        uuid="employee_uuid",
    )

    # Act

    # Create an employee who is a manager (job_pos_id is 1030, 1035 or 1040)
    sd.create_employee(
        {
            "PersonCivilRegistrationIdentifier": cpr_no,
            "Employment": [
                {
                    "EmploymentDate": "2022-06-01",
                    "AnniversaryDate": "2004-08-15",
                    "Profession": {
                        "JobPositionIdentifier": "1030",
                        "ActivationDate": "2022-01-01",
                        "DeactivationDate": "2023-12-31",
                    },
                    "EmploymentStatus": {
                        "EmploymentStatusCode": "1",
                        "ActivationDate": "1970-01-01",
                        "DeactivationDate": "9999-12-31",
                    },
                    "EmploymentIdentifier": "TEST123",
                    "WorkingTime": {"OccupationRate": 1},
                    "EmploymentDepartment": {
                        "DepartmentUUIDIdentifier": org_unit_uuid,
                    },
                }
            ],
        }
    )

    # Assert

    details = sd.importer.employee_details[cpr_no]
    engagement, manager = details

    assert len(details) == 2
    assert engagement.type_id == "engagement"
    assert manager.type_id == "manager"

    assert manager.date_from == "2022-06-01"
    assert manager.date_to == "2023-12-31"


@patch("sdlon.sd_importer.uuid.uuid4")
def test_create_historic_dummy_engagement(mock_uuid4):
    """
    Test that a historic dummy engagement is created if the employee
    engagement start date is older than the corresponding org unit
    start date (see https://redmine.magenta-aps.dk/issues/51898).
    """

    # Arrange

    mock_uuid4.side_effect = ["historic_org_unit_uuid", "engagement_uuid"]
    sd = get_sd_importer()
    org_unit_uuid = str(uuid4())

    sd.nodes[org_unit_uuid] = attrdict(
        {
            "name": "org_unit",
            "date_from": "2000-01-01",  # Later than 1970-01-01 (see below)
        }
    )

    sd.importer.organisation_units[org_unit_uuid] = OrganisationUnitType(
        name="org_unit",
        type_ref="type_ref",
        date_from="2000-01-01",  # Later than 1970-01-01 (see below)
    )

    cpr_no = "0101709999"
    sd.importer.add_employee(
        name=("given_name", "sur_name"),
        identifier=cpr_no,
        cpr_no=cpr_no,
        user_key="employee_user_key",
        uuid="employee_uuid",
    )

    # Act

    # Create an employee on leave (SD EmploymentStatusCode = 3)
    sd.create_employee(
        {
            "PersonCivilRegistrationIdentifier": cpr_no,
            "Employment": [
                {
                    "EmploymentDate": "1960-01-01",
                    "AnniversaryDate": "2004-08-15",
                    "Profession": {"JobPositionIdentifier": "job_id_123"},
                    "EmploymentStatus": {
                        "EmploymentStatusCode": "1",
                        "ActivationDate": "1970-01-01",
                        "DeactivationDate": "9999-12-31",
                    },
                    "EmploymentIdentifier": "TEST123",
                    "WorkingTime": {"OccupationRate": 1},
                    "EmploymentDepartment": {
                        "DepartmentUUIDIdentifier": org_unit_uuid,
                    },
                }
            ],
        }
    )

    # Assert

    details = sd.importer.employee_details[cpr_no]
    engagement, engagement_historic_dummy = details

    assert mock_uuid4.call_count == 2

    # Two engagements are expected - one active and one historic dummy
    assert len(details) == 2

    assert engagement.type_id == "engagement"
    assert engagement_historic_dummy.type_id == "engagement"
    assert engagement.date_from == "2000-01-01"
    assert engagement.date_to is None

    assert engagement_historic_dummy.date_from == "1970-01-01"
    assert engagement_historic_dummy.date_to == "1999-12-31"
    assert engagement_historic_dummy.org_unit_ref == "historic_org_unit_uuid"
    assert engagement_historic_dummy.type_ref == "historisk"


@pytest.mark.parametrize("employment_status", ("7", "8", "9"))
def test_skip_creation_of_sd_let_go_employments(employment_status: str) -> None:
    """
    Test that engagements are not created for SD employees with an SD "let go"
    (i.e. retired etc.) employment status
    """
    # Arrange
    sd = get_sd_importer()
    sd.importer = MagicMock()

    person = {
        "PersonCivilRegistrationIdentifier": "1234567890",
        "Employment": [
            {"EmploymentStatus": {"EmploymentStatusCode": employment_status}}
        ],
    }

    # Act
    sd.create_employee(person)

    # Assert
    sd.importer.add_engagement.assert_not_called()


def test_employment_date_as_engagement_start_date_disabled_per_default():
    sd = get_sd_importer()
    assert sd.employment_date_as_engagement_start_date is False


@parameterized.expand(
    [
        (
            # Test that the employee is moved from an "Afdelings-niveau" to
            # the above "NY1-niveau"
            ["Afdelings-niveau"],
            UUID("00000000-0000-0000-0000-000000000000"),
            UUID("10000000-0000-0000-0000-000000000000"),
        ),
        (
            # Test that the employee is moved from an "Afdelings-niveau" to
            # the "NY2-niveau" two levels above
            ["NY1-niveau", "Afdelings-niveau"],
            UUID("00000000-0000-0000-0000-000000000000"),
            UUID("20000000-0000-0000-0000-000000000000"),
        ),
    ]
)
def test_get_employee_target_unit_uuid(
    too_deep: list[str], original_unit_uuid: UUID, expected: UUID
) -> None:
    # Arrange
    sd = get_sd_importer()

    ny2_uuid = UUID("20000000-0000-0000-0000-000000000000")
    ny1_uuid = UUID("10000000-0000-0000-0000-000000000000")
    afd_uuid = UUID("00000000-0000-0000-0000-000000000000")

    ny2_node = Node("NY2-niveau", uuid=str(ny2_uuid))
    ny1_node = Node("NY1-niveau", uuid=str(ny1_uuid), parent=ny2_node)
    afd_node = Node("Afdelings-niveau", uuid=str(afd_uuid), parent=ny1_node)

    sd.nodes = {
        str(ny2_uuid): ny2_node,
        str(ny1_uuid): ny1_node,
        str(afd_uuid): afd_node,
    }

    # Act
    target_unit_uuid = sd._get_employee_target_unit_uuid(too_deep, original_unit_uuid)

    # Assert
    assert target_unit_uuid == expected


def test_get_employee_target_unit_uuid_afd_is_top_level() -> None:
    """
    Test that the MO target unit is the original "Afdelings-niveau" unit if
    the unit is located directly below the SD institution.
    """

    # Arrange
    sd = get_sd_importer()

    afd_uuid = UUID("00000000-0000-0000-0000-000000000000")
    afd_node = Node("Afdelings-niveau", uuid=str(afd_uuid))

    sd.nodes = {str(afd_uuid): afd_node}

    # Act
    target_unit_uuid = sd._get_employee_target_unit_uuid(["Afdelings-niveau"], afd_uuid)

    # Assert
    assert target_unit_uuid == afd_uuid



from uuid import UUID

import click

from sdlon.sd_common import generate_uuid

def _get_class_uuid(department_level_identifier: str) -> UUID:
    """
    Generate org_unit_level class UUID according to the way it is
    done by the SD-importer. The UUID is necessary when manually creating
    new org_unit_levels after the initial SD-import is done.

    Args:
        department_level_identifier: the SD DepartmentLevelIdentifier for
          the org_unit_level, e.g. "NY7-niveau"

    Returns:
        UUID that the org_unit_level class has to be created with
    """
    class_uuid: str = generate_uuid(
        department_level_identifier,
        None,
        "Silkeborg Kommune"
    )
    return UUID(class_uuid)


@click.command()
# The SD DepartmentLevelIdentifier for the department
@click.argument("department_level_identifier")
def get_class_uuid(department_level_identifier: str) -> None:
    click.echo(_get_class_uuid(department_level_identifier))


if __name__ == "__main__":
    get_class_uuid()



import asyncio
import logging
from datetime import date
from functools import lru_cache
from operator import itemgetter
from typing import cast
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import aiohttp.client_exceptions
import config
from aiohttp import ClientResponse
from aiohttp import ClientSession
from aiohttp import ClientTimeout
from aiohttp import TCPConnector
from more_itertools import chunked
from mox_helpers.mox_helper import create_mox_helper
from mox_helpers.mox_helper import MoxHelper
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.headers import TokenSettings
from ra_utils.tqdm_wrapper import tqdm
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_random_exponential

# The number of retry attempts seems to mostly be 7 various places in DIPEX
retry_attempts = 7


logger = logging.getLogger(__name__)


def get_tcp_connector():
    settings = config.get_config()
    return TCPConnector(limit=settings.max_concurrent_requests)


def get_client_session():
    return ClientSession(
        connector=get_tcp_connector(), timeout=ClientTimeout(total=None)
    )


async def create_details(
    session: ClientSession, detail_payloads: Iterable[dict]
) -> None:
    """Helper function for submitting create detail payloads"""
    url = "/service/details/create"
    await submit_payloads(session, url, detail_payloads, "create details")


async def edit_details(session: ClientSession, detail_payloads: Iterable[dict]) -> None:
    """Helper function for submitting edit detail payloads"""
    url = "/service/details/edit"
    await submit_payloads(session, url, detail_payloads, "edit details")


async def terminate_details(
    session: ClientSession,
    detail_payloads: Iterable[dict],
    ignored_http_statuses: Optional[Tuple[int]] = (404,),
) -> None:
    """Helper function for submitting terminate detail payloads"""
    url = "/service/details/terminate"
    await submit_payloads(
        session,
        url,
        detail_payloads,
        "terminate details",
        ignored_http_statuses=ignored_http_statuses,
    )


async def create_it(payload: dict, obj_uuid: str, mox_helper: MoxHelper) -> None:
    """Helper function for reating an IT system"""
    await mox_helper.insert_organisation_itsystem(payload, obj_uuid)


async def create_klasse(payload: dict, obj_uuid: str, mox_helper: MoxHelper) -> None:
    """Helper function for creating a Klasse object"""
    await mox_helper.insert_klassifikation_klasse(payload, obj_uuid)


async def raise_on_unhandled_mo_error(
    endpoint: str,
    doc: Iterable[dict],
    response: ClientResponse,
    ignored_mo_error_keys: Tuple[str] = ("V_DUPLICATED_IT_USER",),
) -> None:
    """Raise an exception on any MO errors that we do not know to handle.

    :param endpoint: The MO endpoint that was called
    :param doc: The payload used when calling the MO endpoint
    :param response: The MO response returned by the endpoint
    :param ignored_mo_error_keys: Zero or more MO error keys to ignore, if encountered
    """
    if not response.ok:
        # Print out the MO error response to ease debugging
        mo_error = await response.json()
        logger.error(
            "Posting '%s' to '%s' resulted in MO error response '%s'",
            doc,
            endpoint,
            mo_error,
        )
        # Examine the MO error to see if it should be ignored
        mo_error_key = mo_error.get("error_key")
        needs_check = mo_error_key and ignored_mo_error_keys
        if needs_check and mo_error_key in ignored_mo_error_keys:
            logger.debug("Continuing on ignored MO error ''%s", mo_error_key)
            return

    response.raise_for_status()


@retry(
    retry=retry_if_exception_type(aiohttp.ClientConnectionError),
    wait=wait_random_exponential(multiplier=1, max=5),
    stop=stop_after_attempt(retry_attempts),
    reraise=True,
)
async def submit_payloads(
    session: ClientSession,
    endpoint: str,
    payloads: Iterable[dict],
    description: str,
    ignored_http_statuses: Optional[Tuple[int]] = None,
) -> None:
    """
    Send a list of payloads to OS2mo. The payloads are chunked based on preset variable
    and submitted concurrently.

    :param session: A aiohttp session
    :param endpoint: Which endpoint to send the payloads to
    :param payloads: An iterable of dict payloads
    :param description: A description to print as part of the output
    :param ignored_http_statuses: A tuple of HTTP status codes to ignore, if encountered
    """
    settings = config.get_config()
    base_url = settings.mora_base
    headers = TokenSettings().get_headers()

    async def submit(data: List[dict]) -> None:
        doc = list(data)
        # Use semaphore to throttle the amount of concurrent requests
        async with session.post(
            base_url + endpoint,
            params={"force": 1},
            json=doc,
            headers=headers,
        ) as response:
            if ignored_http_statuses and response.status in ignored_http_statuses:
                logger.debug(
                    "%s returned status %r, ignoring",
                    endpoint,
                    response.status,
                )
            else:
                await raise_on_unhandled_mo_error(endpoint, doc, response)

    chunks = chunked(payloads, settings.os2mo_chunk_size)
    tasks = list(map(submit, chunks))
    if len(tasks) == 0:
        return

    for f in tqdm(
        asyncio.as_completed(tasks), total=len(tasks), unit="chunk", desc=description
    ):
        await f


async def lookup_organisationfunktion():
    """Helper function for fetching all available 'organisationfunktion' objects."""
    settings = config.get_config()
    mox = await create_mox_helper(settings.mox_base)
    return await mox.search_organisation_organisationfunktion(params={"bvn": "%"})


def lookup_employees():
    """Helper function for fetching all available 'employee' objects."""
    settings = config.get_config()
    mh = MoraHelper(hostname=settings.mora_base, export_ansi=True)
    return mh.read_all_users()


@lru_cache(maxsize=0)
def build_cpr_map() -> Dict[str, str]:
    employees = lookup_employees()
    cache = cast(
        Iterator[Tuple[str, str]],
        map(
            itemgetter("cpr_no", "uuid"),
            filter(lambda employee: employee.get("cpr_no"), employees),
        ),
    )
    return dict(cache)


def convert_validities(from_time: date, to_time: date) -> Tuple[str, Optional[str]]:
    from_time_str = from_time.isoformat()
    to_time_str = to_time.isoformat()
    return from_time_str, to_time_str if to_time_str != "9999-12-31" else None



import datetime

from integrations.calculate_primary.common import logger
from integrations.calculate_primary.common import MOPrimaryEngagementUpdater


class DefaultPrimaryEngagementUpdater(MOPrimaryEngagementUpdater):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def remove_past(user_uuid, no_past, eng):
            if no_past and eng["validity"]["to"]:
                to = datetime.datetime.strptime(eng["validity"]["to"], "%Y-%m-%d")
                if to < datetime.datetime.now():
                    return False
            return True

        self.check_filters = []
        self.calculate_filters = [remove_past]

    def _find_primary_types(self):
        """
        Read the engagement types from MO and match them up against the three
        known types in the OPUS->MO import.
        :param helper: An instance of mora-helpers.
        :return: A dict matching up the engagement types with LoRa class uuids.
        """
        PRIMARY = "primary"
        NON_PRIMARY = "non-primary"
        FIXED_PRIMARY = "explicitly-primary"

        logger.info("Read primary types")
        primary_dict = {"fixed_primary": None, "primary": None, "non_primary": None}

        primary_types = self.helper.read_classes_in_facet("primary_type")
        for primary_type in primary_types[0]:
            if primary_type["user_key"] == PRIMARY:
                primary_dict["primary"] = primary_type["uuid"]
            if primary_type["user_key"] == NON_PRIMARY:
                primary_dict["non_primary"] = primary_type["uuid"]
            if primary_type["user_key"] == FIXED_PRIMARY:
                primary_dict["fixed_primary"] = primary_type["uuid"]

        if None in primary_dict.values():
            raise Exception("Missing primary types: {}".format(primary_dict))
        primary_list = [primary_dict["fixed_primary"], primary_dict["primary"]]

        return primary_dict, primary_list

    def _find_primary(self, mo_engagements):
        if not mo_engagements:
            return None

        # The primary engagement is the engagement with the highest occupation rate.
        # - The occupation rate is found as 'fraction' on the engagement.
        #
        # If two engagements have the same occupation rate, the tie is broken by
        # picking by user-key.
        primary_engagement = max(
            mo_engagements,
            # Sort first by fraction, then reversely by user_key integer
            key=lambda eng: (eng.get("fraction") or 0, eng["user_key"]),
        )
        return primary_engagement["uuid"]



from typing import OrderedDict, Any

from sdlon.models import EmploymentWithTelephoneNumberIdentifier, SDBasePerson
from sdlon.sd_common import ensure_list


def convert_to_sd_base_person(person: OrderedDict[str, Any]) -> SDBasePerson:
    """
    Convert the raw OrderedDict from SD to a Pydantic base model
    representing the SD person.

    Args:
        person: the person from SD Løn

    Returns:
        Basic SD person
    """

    telephone_number_identifiers = []
    employments = ensure_list(person.get("Employment", []))
    for employment in employments:
        contact_information = employment.get("ContactInformation", {})
        telephone_number_ids = ensure_list(
            contact_information.get("TelephoneNumberIdentifier", [])
        )
        telephone_number_identifiers.append(
            EmploymentWithTelephoneNumberIdentifier(
                employment_identifier=employment["EmploymentIdentifier"],
                telephone_number_ids=telephone_number_ids,
            )
        )

    return SDBasePerson(
        cpr=person["PersonCivilRegistrationIdentifier"],
        given_name=person.get("PersonGivenName"),
        surname=person.get("PersonSurnameName"),
        emp_with_telephone_number_identifiers=telephone_number_identifiers,
    )



from uuid import uuid4

import initial
import pytest
from more_itertools import first_true
from uuids import AZID_SYSTEM

from .helpers import mock_config
from .helpers import mock_create_mox_helper


@pytest.mark.asyncio
async def test_import_it_does_nothing_if_customized():
    with mock_config(azid_it_system_uuid=uuid4()), mock_create_mox_helper() as mh:
        await initial.import_it()
        mh.assert_not_awaited()


@pytest.mark.asyncio
async def test_import_it_creates_it_system_if_not_customised():
    with mock_config(azid_it_system_uuid=AZID_SYSTEM), mock_create_mox_helper() as mh:
        await initial.import_it()
        mh.assert_awaited_once()


@pytest.mark.asyncio
async def test_import_remaining_classes():
    with mock_config(), mock_create_mox_helper():
        result = await initial.import_remaining_classes()
        non_primary_class = first_true(
            result, pred=lambda model: model.source.bvn == "non-primary"
        )
        assert non_primary_class is not None



from datetime import datetime
from typing import Tuple
from unittest.mock import Mock
from unittest.mock import patch
from uuid import UUID
from uuid import uuid4

import los_pers
from hypothesis import given
from hypothesis import settings
from hypothesis import strategies as st

from .helpers import HelperMixin
from .helpers import mock_config


_azid_it_system_uuid = uuid4()
_mock_config = mock_config(azid_it_system_uuid=_azid_it_system_uuid)
_primary_strategy = st.one_of(st.just("Ja"), st.just("Nej"))


@patch("util.build_cpr_map", return_value={})
class TestPersonImporter:
    @given(st.from_type(los_pers.Person), _primary_strategy)
    def test_generate_employee_az_id_payload_uses_setting(
        self,
        mock_build_cpr_map: Mock,
        person: los_pers.Person,
        primary: str,
    ):
        person.primary = primary
        with _mock_config:
            importer = los_pers.PersonImporter()
        payload = importer.generate_employee_az_id_payload(person, "2021-11-17")
        assert payload["itsystem"]["uuid"] == str(_azid_it_system_uuid)

    @given(st.from_type(los_pers.Person), _primary_strategy)
    def test_ad_accounts_are_not_created(
        self,
        mock_build_cpr_map: Mock,
        person: los_pers.Person,
        primary: str,
    ):
        person.primary = primary
        with _mock_config:
            importer = los_pers.PersonImporter()
        detail_payloads = importer.create_detail_payloads([person], None)

        # Verify that we only generate "AZ" IT users (not "AD")
        it_user_payloads = [
            payload for payload in detail_payloads if payload["type"] == "it"
        ]
        if person.az_id:
            assert len(it_user_payloads) == 1
            assert it_user_payloads[0]["user_key"] == person.az_id
            assert it_user_payloads[0]["itsystem"]["uuid"] == str(_azid_it_system_uuid)
        else:
            assert len(it_user_payloads) == 0


class TestPersonImporterReusesCPR(HelperMixin):
    @given(st.booleans(), st.from_type(los_pers.Person))
    def test_get_person_mo_uuid(self, present: bool, person: los_pers.Person):
        """If the person's CPR is already in MO (present=True), reuse the existing
        person UUID.
        If the person's CPR is not already in MO (present=False), create a new UUID
        derived from the person's CPR.
        """
        # Arrange
        importer, expected_person_uuid = self._get_importer_and_expected_person_uuid(
            present, person
        )
        # Act
        actual_person_uuid = importer._get_person_mo_uuid(person)
        # Assert
        assert actual_person_uuid == expected_person_uuid

    @settings(deadline=None)
    @given(
        st.booleans(), st.from_type(los_pers.Person), _primary_strategy, st.datetimes()
    )
    def test_handle_create_uses_correct_person_uuid(
        self, present: bool, person: los_pers.Person, primary: str, filedate: datetime
    ):
        # Arrange
        person.primary = primary
        importer, expected_person_uuid = self._get_importer_and_expected_person_uuid(
            present, person
        )

        # Act: run `handle_create`
        mock_csv = self._mock_read_csv(person)
        mock_session = self._mock_get_client_session()
        with mock_csv, mock_session:
            with self._mock_create_details() as mock_create_details:
                self._run_until_complete(
                    importer.handle_create("unused_filename.csv", filedate)
                )

        # Assert that we used the proper person UUID in the "create payloads" created
        self._assert_correct_person_uuid_in_create_payloads(
            mock_create_details, expected_person_uuid
        )

    @given(
        st.booleans(), st.from_type(los_pers.Person), _primary_strategy, st.datetimes()
    )
    def test_handle_edit_uses_correct_person_uuid(
        self, present: bool, person: los_pers.Person, primary: str, filedate: datetime
    ):
        # Arrange
        person.primary = primary
        importer, expected_person_uuid = self._get_importer_and_expected_person_uuid(
            present, person
        )

        # Act: run `handle_edit`
        mock_csv = self._mock_read_csv(person)
        mock_session = self._mock_get_client_session()
        with mock_csv, mock_session:
            with self._mock_create_details() as mock_create_details:
                with self._mock_edit_details() as mock_edit_details:
                    # Generate an address UUID in the same manner as
                    # `PersonImporter.generate_employee_email_payload`
                    orgfunk_uuid = importer.uuid_generator(person.cpr + "email")
                    with self._mock_lookup_organisationfunktion(
                        return_value={orgfunk_uuid}
                    ):
                        self._run_until_complete(
                            importer.handle_edit("unused_filename.csv", filedate)
                        )

        # Assert that we used the proper person UUID in the "create payloads" created
        self._assert_correct_person_uuid_in_create_payloads(
            mock_create_details, expected_person_uuid
        )
        # Assert that we used the proper person UUID in the "edit payloads" created
        self._assert_correct_person_uuid_in_edit_payloads(
            mock_edit_details, expected_person_uuid
        )

    def _get_importer_and_expected_person_uuid(
        self, present: bool, person: los_pers.Person
    ) -> Tuple[los_pers.PersonImporter, UUID]:
        expected_person_uuid = uuid4()
        cpr_map = {person.cpr: expected_person_uuid} if present else {}

        with patch("util.build_cpr_map", return_value=cpr_map):
            with _mock_config:
                importer = los_pers.PersonImporter()
                if not present:
                    expected_person_uuid = importer.uuid_generator(person.cpr)

        return importer, expected_person_uuid

    def _assert_correct_person_uuid_in_create_payloads(
        self, mock_create_details: Mock, expected_person_uuid: UUID
    ):
        for payload in mock_create_details.call_args[0][1]:
            if payload["type"] == "employee":
                assert UUID(payload["uuid"]) == expected_person_uuid
            else:
                assert UUID(payload["person"]["uuid"]) == expected_person_uuid

    def _assert_correct_person_uuid_in_edit_payloads(
        self, mock_edit_details: Mock, expected_person_uuid: UUID
    ):
        for payload in mock_edit_details.call_args[0][1]:
            if payload["type"] == "employee":
                assert UUID(payload["data"]["uuid"]) == expected_person_uuid
            else:
                assert UUID(payload["data"]["person"]["uuid"]) == expected_person_uuid



import json
from datetime import datetime
from itertools import starmap
from operator import itemgetter

import click
from more_itertools import flatten
from os2mo_helpers.mora_helpers import MoraHelper

from .sd_common import EmploymentStatus
from .sd_common import load_settings
from .sd_common import sd_lookup


def progress_iterator(elements, outputter, mod=10):
    total = len(elements)
    for i, element in enumerate(elements, start=1):
        if i == 1 or i % mod == 0 or i == total:
            outputter("{}/{}".format(i, total))
        yield element


class TestMOAgainstSd(object):
    def __init__(self):
        self.settings = load_settings()
        self.date = datetime.now()

        self.helper = MoraHelper(hostname=self.settings["mora.base"], use_cache=False)

    def _compare_dates(self, sd_employment, mo_engagement):
        """Check dates for discrepancies."""
        status = EmploymentStatus(
            sd_employment["EmploymentStatus"]["EmploymentStatusCode"]
        )
        # TODO, we should also consider to check for actual date_values.
        # This would need a loop that finds the entire validity of the engagement
        if status in EmploymentStatus.on_payroll():
            # Check that MO agrees that the engagement is active
            is_ok = (
                mo_engagement["validity"]["from"]
                < self.date
                < mo_engagement["validity"]["to"]
            )
            # If false, should be current
            return "dates_current", is_ok
        if status in EmploymentStatus.let_go():
            # employment_end_date = sd_employment['EmploymentStatus']['ActivationDate']
            is_ok = mo_engagement["validity"]["to"] < self.date
            # If false, should have been terminated
            return "dates_past", is_ok
        if status == EmploymentStatus.AnsatUdenLoen:
            return "status0", True
        msg = "Unhandled status, fix this program: " + str(status)
        raise NotImplementedError(msg)

    def _compare_salary(self, sd_employment, mo_engagement):
        """Check salary status for discrepancies."""
        mo_no_salary = mo_engagement["primary"]["user_key"] == "status0"
        sd_status = EmploymentStatus(
            sd_employment["EmploymentStatus"]["EmploymentStatusCode"]
        )
        sd_no_salary = sd_status == EmploymentStatus.AnsatUdenLoen
        is_ok = mo_no_salary == sd_no_salary
        return "salary", is_ok

    def _compare_job_function(self, sd_employment, mo_engagement):
        """Check job function for discrepancies."""
        mo_job_function_id = mo_engagement["job_function"]["user_key"]
        sd_job_position_id = sd_employment["Profession"]["JobPositionIdentifier"]
        is_ok = mo_job_function_id == sd_job_position_id
        return "job_function", is_ok

    def _compare_employments(self, sd_employment, mo_engagement):
        """Check a single employment pair for discrepancies.

        Returns a tuple; id, dictionary on the following form:

        employment_id, {
            "check_name": False
            ...
        }
        """
        employment_id = sd_employment["EmploymentIdentifier"]
        if mo_engagement is None:
            return employment_id, {"no_match": False}

        status_checks = dict(
            [
                self._compare_dates(sd_employment, mo_engagement),
                self._compare_salary(sd_employment, mo_engagement),
                self._compare_job_function(sd_employment, mo_engagement),
            ]
        )
        return employment_id, status_checks

    def check_user(self, mo_uuid):
        """Check a single employee for discrepancies.

        Returns a tuple; uuid, dictionary on the following form:

        user_uuid, {
            'uuid': user_uuid,
            'employments': {
                'employment_id': {
                    "check_name": False,
                    ...
                },
                ...
            },
            'all_ok': False # Boolean for whether all employments were all ok
        }
        """
        mo_user = self.helper.read_user(user_uuid=mo_uuid)
        mo_engagements = self.helper.read_user_engagement(user=mo_uuid, read_all=True)
        # List all distinct engagements, collapsed to one line pr. engagement
        condensed_engagements = {}

        for eng in mo_engagements:
            if eng["validity"]["to"] is None:
                eng["validity"]["to"] = "9999-12-31"
            eng["validity"]["from"] = datetime.strptime(
                eng["validity"]["from"], "%Y-%m-%d"
            )
            eng["validity"]["to"] = datetime.strptime(eng["validity"]["to"], "%Y-%m-%d")

            user_key = eng["user_key"]
            # from_date = datetime.strptime(eng['validity']['from'], '%Y-%m-%d')
            # to_date = datetime.strptime(eng['validity']['to'], '%Y-%m-%d')
            from_date = eng["validity"]["from"]
            to_date = eng["validity"]["to"]
            if user_key in condensed_engagements:
                if from_date < condensed_engagements[user_key]["validity"]["from"]:
                    condensed_engagements[user_key]["validity"]["from"] = from_date
                if to_date > condensed_engagements[user_key]["validity"]["to"]:
                    condensed_engagements[user_key]["validity"]["to"] = to_date
            else:
                condensed_engagements[user_key] = eng

        # Notice, this will not get future engagements
        params = {
            "PersonCivilRegistrationIdentifier": mo_user["cpr_no"],
            "StatusActiveIndicator": "true",
            "StatusPassiveIndicator": "true",
            "DepartmentIndicator": "true",
            "EmploymentStatusIndicator": "true",
            "ProfessionIndicator": "true",
            "WorkingTimeIndicator": "true",
            "UUIDIndicator": "true",
            "SalaryAgreementIndicator": "false",
            "SalaryCodeGroupIndicator": "false",
            "EffectiveDate": self.date.strftime("%d.%m.%Y"),
        }
        sd_employments_response = sd_lookup("GetEmployment20111201", params)
        if "Person" not in sd_employments_response:
            return mo_uuid, {
                "uuid": mo_uuid,
                "employments": {"Person not found in SD": {}},
                "all_ok": False,
            }
        employments = sd_employments_response["Person"]["Employment"]
        if not isinstance(employments, list):
            employments = [employments]

        def create_mo_pair(sd_employment):
            employment_id = sd_employment["EmploymentIdentifier"]
            return sd_employment, condensed_engagements.get(employment_id)

        # NOTE: Currently we only find entries in SD that are not in MO,
        #       Not entries in MO that are not in SD
        pairs = map(create_mo_pair, employments)
        # check_status has the format of employment_id --> dict,
        # where the resulting dict has the format of check_name --> bool,
        # where the boolean is whether the check passed or not
        check_status = dict(starmap(self._compare_employments, pairs))
        return mo_uuid, {
            "uuid": mo_uuid,
            "employments": check_status,
            "all_ok": all(
                # collapse the check_status to a list of booleans
                flatten(checks.values() for checks in check_status.values())
            ),
        }

    def check_department(self, department_uuid):
        """Check all employees in an organisation for discrepancies.

        Returns a tuple; uuid, dictionary on the following form:

        department_uuid, {
            'uuid': department_uuid,
            'users': {
                'user_uuid': {
                    'uuid': user_uuid,
                    'employments': {
                        'employment_id': {
                            "check_name": False,
                            ...
                        },
                        ...
                    },
                    'all_ok': False
                },
                ...
            },
            'all_ok': False # Boolean for whether all users were all ok
        }
        """
        employees = self.helper.read_organisation_people(
            department_uuid, read_all=True
        ).keys()
        users = dict(map(self.check_user, employees))
        return department_uuid, {
            "uuid": department_uuid,
            "users": users,
            "all_ok": all(user["all_ok"] for user in users.values()),
        }

    def check_all(self, limit, progress):
        """Check limit/all employees in an MO instance for discrepancies.

        Returns a dictionary on the following form:

        {
            'users': {
                'user_uuid': {
                    'uuid': user_uuid,
                    'employments': {
                        'employment_id': {
                            "check_name": False,
                            ...
                        },
                        ...
                    },
                    'all_ok': False
                },
                ...
            },
            'all_ok': False # Boolean for whether all users were all ok
        }
        """
        employees = self.helper.read_all_users(limit=limit)
        if progress:
            employees = progress_iterator(employees, print)
        employees = map(itemgetter("uuid"), employees)
        users = dict(map(self.check_user, employees))
        return {
            "users": users,
            "all_ok": all(user["all_ok"] for user in users.values()),
        }

    def _print_status(self, name, status):
        print(name + ":", " " * (20 - len(name)), "\u2713" if status else "x")

    def print_user_result(self, test_result):
        """Pretty-print a result from check_user."""
        print("*" * 55)
        print("Checking employee: {}".format(test_result["uuid"]))
        for engagement_id, checks in sorted(test_result["employments"].items()):
            print("Checking engagement: {}".format(engagement_id))
            for key, status_check in sorted(checks.items()):
                self._print_status(key, status_check)
            print()
        print("-" * 26)
        self._print_status("all_ok", test_result["all_ok"])
        print("*" * 55)

    def print_all_result(self, test_result):
        for user_uuid, user_result in sorted(test_result["users"].items()):
            self.print_user_result(user_result)
        print()
        self._print_status("total_all_ok", test_result["all_ok"])

    def print_department_result(self, test_result):
        print("Checking department {}".format(test_result["uuid"]))
        print()
        self.print_all_result(test_result)


@click.group()
@click.option("--json/--no-json", default=False, help="Output as JSON.")
@click.option("--progress/--no-progress", default=False, help="Print progress.")
@click.pass_context
def cli(ctx, json, progress):
    """MO to SD equivalence testing utility.

    Compares MO against SD to find discrepancies in the data.
    """
    # ensure that ctx.obj exists and is a dict, no matter how it is called.
    ctx.ensure_object(dict)
    ctx.obj["json"] = json
    ctx.obj["progress"] = progress


@cli.command()
@click.option(
    "--uuid", type=click.UUID, help="UUID of the user to check.", required=True
)
@click.pass_context
# Example UUID: 'fadfcc38-5d42-4857-a950-0adc65babb13'
def check_user(ctx, uuid):
    """Check a single employee."""
    tester = TestMOAgainstSd()
    _, test_result = tester.check_user(str(uuid))

    if ctx.obj["json"]:
        print(json.dumps(test_result, indent=4, sort_keys=True))
    else:
        tester.print_user_result(test_result)


@cli.command()
@click.option(
    "--uuid", type=click.UUID, help="UUID of the organisation to check.", required=True
)
@click.pass_context
# Example UUID: 'ea5a237a-8f8b-4300-9a00-000006180002'
def check_department(ctx, uuid):
    """Check all employees in an organisation."""
    tester = TestMOAgainstSd()
    _, test_result = tester.check_department(str(uuid), ctx.obj["progress"])

    if ctx.obj["json"]:
        print(json.dumps(test_result, indent=4, sort_keys=True))
    else:
        tester.print_department_result(test_result)


@cli.command()
@click.option(
    "--limit",
    type=click.INT,
    help="Number of employees to check. 0 --> All.",
    default=5,
)
@click.pass_context
def check_all(ctx, limit):
    """Check limit/all employees in an MO instance."""
    tester = TestMOAgainstSd()
    test_result = tester.check_all(limit, ctx.obj["progress"])

    if ctx.obj["json"]:
        print(json.dumps(test_result, indent=4, sort_keys=True))
    else:
        tester.print_all_result(test_result)


if __name__ == "__main__":
    cli()



import logging
from abc import ABC
from abc import abstractmethod
from enum import Enum

import requests
from gql import gql
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.headers import TokenSettings
from ra_utils.job_settings import JobSettings
from raclients.graph.client import GraphQLClient

logger = logging.getLogger(__name__)


class Settings(JobSettings):
    integrations_kle_xlsx_file_path: str
    integrations_kle_xlsx_org_unit_levels: list | None = None


class Aspects(Enum):
    Indsigt = 1
    Udfoerende = 2
    Ansvarlig = 3


# Maps between the enum and scopes on the classes from the aspect facet
ASPECT_MAP = {
    "INDSIGT": Aspects.Indsigt,
    "UDFOERENDE": Aspects.Udfoerende,
    "ANSVARLIG": Aspects.Ansvarlig,
}


class KLEAnnotationIntegration(ABC):
    """Import and export of KLE annotation from or to an external source."""

    # XXX: This uses a simple inheritance based pattern. We might want to use
    # something like a Strategy here. However, maybe YAGNI.

    def __init__(self):

        self.settings = Settings()
        self.settings.start_logging_based_on_settings()

        self.mora_base = self.settings.mora_base
        self.mora_session = self._get_mora_session()
        self.helper = MoraHelper(hostname=self.mora_base)
        self.org_uuid = self.helper.read_organisation()

        kle_classes = self.get_kle_classes_from_mo()
        self.kle_uuid_map = {item["user_key"]: item["uuid"] for item in kle_classes}

        aspect_classes = self.get_aspect_classes_from_mo()
        self.aspect_map = {
            ASPECT_MAP[clazz["scope"]]: clazz["uuid"] for clazz in aspect_classes
        }
        self.gql_client = GraphQLClient(
            url=f"{self.mora_base}/graphql/v7",
            client_id=self.settings.client_id,
            client_secret=self.settings.client_secret,
            auth_realm=self.settings.auth_realm,
            auth_server=self.settings.auth_server,
            sync=True,
        )

    def _get_mora_session(self) -> requests.Session:
        s = requests.Session()
        session_headers = TokenSettings().get_headers()

        if session_headers:
            s.headers.update(session_headers)
        s.verify = False
        return s

    def _get_mo_org_uuid(self) -> str:
        """
        Get the UUID of the organisation configured in OS2mo
        :return:
        """
        logger.info("Fetching Organisation UUID from OS2mo")
        r = self.mora_session.get("{}/service/o/".format(self.mora_base))
        r.raise_for_status()
        return r.json()[0]["uuid"]

    def get_kle_classes_from_mo(self) -> list:
        """Get all of the kle_number 'klasse' objects from OS2mo"""
        logger.info("Fetching KLE numbers from OS2mo")
        kle_numbers, _ = self.helper.read_classes_in_facet("kle_number")
        return kle_numbers

    def get_aspect_classes_from_mo(self) -> list:
        """Get all of the kle_aspect 'klasse' objects from OS2mo"""
        logger.info("Fetching KLE aspect classes from OS2mo")
        kle_aspects, _ = self.helper.read_classes_in_facet("kle_aspect")
        return kle_aspects

    def get_all_org_units_from_mo(self) -> list:
        """Get a list of all units from OS2mo"""
        logger.info("Fetching all org units from OS2mo")
        query = gql(
            """
            query OrgUnitQuery {
              org_units {
                objects {
                  current {
                    uuid
                    org_unit_level {
                      name
                    }
                    name
                  }
                }
              }
            }
            """
        )

        r = self.gql_client.execute(query)

        units = [o["current"] for o in r["org_units"]["objects"]]
        # filter by org_unit_level if configured in settings
        if org_unit_levels := self.settings.integrations_kle_xlsx_org_unit_levels:
            units = [
                o for o in units if o["org_unit_level"].get("name") in org_unit_levels
            ]
        logger.info("Found {} units".format(len(units)))
        return units

    def post_payloads_to_mo(self, payloads: list):
        """Submit a list of details payloads to OS2mo"""
        logger.info("Posting payloads to OS2mo ")
        url = "{}/service/details/create".format(self.mora_base)

        r = self.mora_session.post(url, json=payloads, params={"force": 1})
        r.raise_for_status()

    @abstractmethod
    def run(self):
        """Implement this, normally to execute import or export."""
        pass



from uuid import UUID

from gql import gql
from more_itertools import one
from raclients.graph.client import GraphQLClient
from sdlon.models import MOBasePerson

QUERY_GET_EMPLOYEE = gql(
    """
        query GetEmployee($cpr: [CPR!]!) {
            employees(cpr_numbers: $cpr) {
                objects {
                    current {
                        name
                        givenname
                        surname
                        uuid
                    }
                }
            }
        }
    """
)


def get_employee(gql_client: GraphQLClient, cpr: str) -> MOBasePerson | None:
    """
    Get employee from MO

    Args:
        gql_client: The GraphQl client for calling MO
        cpr: The CPR number of the employee

    Returns:
        The MO employee or None if the employee does not exist in MO
    """
    r = gql_client.execute(QUERY_GET_EMPLOYEE, variable_values={"cpr": cpr})

    obj = r["employees"]["objects"]
    if not obj:
        return None

    employee = one(obj)["current"]

    return MOBasePerson(
        cpr=cpr,
        givenname=employee["givenname"],
        surname=employee["surname"],
        name=employee["name"],
        uuid=UUID(employee["uuid"]),
    )



#
# Copyright (c) Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import datetime
import os
import sys
import unittest

import psycopg2
import sd_importer  # noqa
from integration_test_helpers import _count  # noqa
from os2mo_data_import import ImportHelper
from os2mo_helpers.mora_helpers import MoraHelper

sys.path.append("../SD_Lon")

sys.path.append("../../os2mo_data_import/long_tests")

MUNICIPALTY_NAME = os.environ.get("MUNICIPALITY_NAME")
MUNICIPALTY_CODE = os.environ.get("MUNICIPALITY_CODE")
MOX_BASE = os.environ.get("MOX_BASE", "http://localhost:5000")
MORA_BASE = os.environ.get("MORA_BASE", "http://localhost:80")
GLOBAL_GET_DATE = datetime.datetime(2019, 6, 13, 0, 0)


class IntegrationDataTests(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.morah = MoraHelper(use_cache=False)

    def _clear_database(self):
        conn = psycopg2.connect(
            user="mox", dbname="mox", host="localhost", password="mox"
        )
        cursor = conn.cursor()

        query = (
            "select relname from pg_class where relkind='r' "
            + "and relname !~ '^(pg_|sql_)';"
        )

        cursor.execute(query)
        for row in cursor.fetchall():
            query = "truncate {} cascade;".format(row[0])
            cursor.execute(query)
        conn.commit()

    def setUp(self):
        self.importer = ImportHelper(
            create_defaults=True,
            mox_base=MOX_BASE,
            mora_base=MORA_BASE,
        )

    @classmethod
    def tearDownClass(self):
        pass

    def test_010_standard_import(self):
        """
        self a standard SD import
        """
        print("Test 010")
        self._clear_database()
        sd = sd_importer.SdImport(
            self.importer,
            MUNICIPALTY_NAME,
            MUNICIPALTY_CODE,
            import_date_from=GLOBAL_GET_DATE,
            org_only=True,
            org_id_prefix=None,
        )

        sd.create_ou_tree(create_orphan_container=True)
        self.importer.import_all()

        count = _count(MOX_BASE)
        print(count)
        self.assertTrue(count["unit_count"] in range(550, 560))

    def test_011_test_kommunaldirektør(self):
        print("Test 011")
        kommunaldirektør = self.morah.read_ou("3e890083-589b-4a00-ba00-000001350001")
        self.assertFalse("error_key" in kommunaldirektør)

    def test_012_test_udgåede(self):
        print("Test 012")
        udgåede = self.morah.read_ou("8df2c488-8038-4b00-a600-000001460003")
        self.assertFalse("error_key" in udgåede)

    def test_013_test_orphan(self):
        print("Test 013")
        org = self.morah.read_organisation()
        top_units = self.morah.read_top_units(org)
        orphans = 0
        for unit in top_units:
            if unit["name"] == "Forældreløse enheder":
                orphans += 1
        self.assertTrue(orphans == 1)

        # Forældreløse, Kommunalbestyrelsen, Kommnaldirektør, udgåede
        self.assertTrue(len(top_units) == 4)

    def test_020_standard_import(self):
        """
        Test a standard SD import, withoug grouping of orphan units
        """
        print("Test 020")
        self._clear_database()
        sd = sd_importer.SdImport(
            self.importer,
            MUNICIPALTY_NAME,
            MUNICIPALTY_CODE,
            import_date_from=GLOBAL_GET_DATE,
            org_only=True,
            org_id_prefix=None,
        )

        sd.create_ou_tree(create_orphan_container=False)

        self.importer.import_all()
        count = _count(MOX_BASE)
        print(count)
        self.assertTrue(count["unit_count"] in range(550, 555))

    def test_021_test_orphan(self):
        print("Test 021")
        org = self.morah.read_organisation()
        top_units = self.morah.read_top_units(org)
        orphans = 0
        for unit in top_units:
            if unit["name"] == "Forældreløse enheder":
                orphans += 1
        self.assertTrue(orphans == 0)

        self.assertTrue(len(top_units) > 6)

    def test_030_import_subtree(self):
        """
        Test an import of a single sub-tree.
        """
        print("Test 030")
        self._clear_database()
        sd = sd_importer.SdImport(
            self.importer,
            MUNICIPALTY_NAME,
            MUNICIPALTY_CODE,
            import_date_from=GLOBAL_GET_DATE,
            org_only=True,
            org_id_prefix=None,
        )

        sd.create_ou_tree(
            create_orphan_container=False,
            # Direktør m.a. CAY,CBK,CTM, CEI
            sub_tree="3e890083-589b-4a00-ba00-000001350001",
        )

        self.importer.import_all()
        count = _count(MOX_BASE)
        print(count)
        self.assertTrue(count["unit_count"] in range(460, 465))

    def test_031_test_roots(self):
        print("Test 031")
        org = self.morah.read_organisation()
        top_units = self.morah.read_top_units(org)
        self.assertTrue(len(top_units) == 1)

    def test_040_import_subtree_with_super(self):
        """
        Test an import of a single sub-tree with a super unit.
        """
        print("Test 040")
        self._clear_database()
        sd = sd_importer.SdImport(
            self.importer,
            MUNICIPALTY_NAME,
            MUNICIPALTY_CODE,
            import_date_from=GLOBAL_GET_DATE,
            org_only=True,
            org_id_prefix=None,
        )

        adm_name = "AdmOrg"
        self.importer.add_klasse(
            identifier=adm_name,
            facet_type_ref="org_unit_type",
            user_key=adm_name,
            scope="TEXT",
            title=adm_name,
        )

        self.importer.add_organisation_unit(
            identifier=adm_name,
            name=adm_name,
            user_key=adm_name,
            type_ref=adm_name,
            date_from="1930-01-01",
            date_to=None,
            parent_ref=None,
        )

        sd.create_ou_tree(
            create_orphan_container=False,
            super_unit=adm_name,
            sub_tree="3e890083-589b-4a00-ba00-000001350001",
        )

        self.importer.import_all()
        count = _count(MOX_BASE)
        print(count)
        self.assertTrue(count["unit_count"] in range(450, 470))

    def test_041_test_top_units(self):
        print("Test 041")
        org = self.morah.read_organisation()
        top_units = self.morah.read_top_units(org)
        super_unit = 0
        for unit in top_units:
            if unit["name"] == "AdmOrg":
                super_unit += 1
        self.assertTrue(super_unit == 1)

        self.assertTrue(len(top_units) == 1)

    def test_050_import_subtree_with_prefix(self):
        """
        Test an import of bvn-prefixed units.
        """
        print("Test 050")
        self._clear_database()
        sd = sd_importer.SdImport(
            self.importer,
            MUNICIPALTY_NAME,
            MUNICIPALTY_CODE,
            import_date_from=GLOBAL_GET_DATE,
            org_only=True,
            org_id_prefix="adm_",
        )

        sd.create_ou_tree(
            create_orphan_container=False,
            # 'Direktionen encoded with org prefix 'adm_'
            sub_tree="fff9e2a6-d670-b656-c719-994eeac03a74",
        )
        self.importer.import_all()

        count = _count(MOX_BASE)
        print(count)
        self.assertTrue(count["unit_count"] in range(450, 470))

    def test_051_test_prefix(self):
        print("Test 510")
        org = self.morah.read_organisation()
        top_units = self.morah.read_top_units(org)
        bvn = top_units[0]["user_key"]
        self.assertTrue(bvn[0:4] == "adm_")

    def test_060_full_double_import(self):
        # Import of Administration
        adm_name = "AdmOrg"
        print("Test 060")
        self._clear_database()

        self.importer.add_klasse(
            identifier=adm_name,
            facet_type_ref="org_unit_type",
            user_key=adm_name,
            scope="TEXT",
            title=adm_name,
        )

        self.importer.add_organisation_unit(
            identifier=adm_name,
            uuid="11111111-1111-1111-1111-111111111111",
            name=adm_name,
            user_key=adm_name,
            type_ref=adm_name,
            date_from="1930-01-01",
            date_to=None,
            parent_ref=None,
        )

        sd = sd_importer.SdImport(
            self.importer,
            MUNICIPALTY_NAME,
            MUNICIPALTY_CODE,
            import_date_from=GLOBAL_GET_DATE,
            org_only=True,
            org_id_prefix="adm_",
        )

        sd.create_ou_tree(
            create_orphan_container=False,
            # 'Direktionen encoded with org prefix 'adm_'
            sub_tree="fff9e2a6-d670-b656-c719-994eeac03a74",
            super_unit=adm_name,
        )

        # Import of Lønorganisation
        løn_name = "LønOrg"

        self.importer.add_klasse(
            identifier=løn_name,
            facet_type_ref="org_unit_type",
            user_key=løn_name,
            scope="TEXT",
            title=løn_name,
        )

        self.importer.add_organisation_unit(
            identifier=løn_name,
            uuid="22222222-2222-2222-2222-222222222222",
            name=løn_name,
            user_key=løn_name,
            type_ref=løn_name,
            date_from="1930-01-01",
            date_to=None,
            parent_ref=None,
        )

        sd = sd_importer.SdImport(
            self.importer,
            MUNICIPALTY_NAME,
            MUNICIPALTY_CODE,
            import_date_from=GLOBAL_GET_DATE,
            org_only=True,
            org_id_prefix=None,
        )

        sd.create_ou_tree(create_orphan_container=False, super_unit=løn_name)
        self.importer.import_all()
        count = _count(MOX_BASE)
        print(count)
        self.assertTrue(count["unit_count"] in range(1000, 1050))

    def test_061_test_top_units(self):
        print("Test 061")
        org = self.morah.read_organisation()
        top_units = self.morah.read_top_units(org)
        self.assertTrue(len(top_units) == 2)

    def test_062_test_kommunaldirektør(self):
        print("Test 062")
        kommunaldirektør = self.morah.read_ou("3e890083-589b-4a00-ba00-000001350001")
        self.assertFalse("error_key" in kommunaldirektør)

    def test_063_test_udgåede(self):
        print("Test 63")
        udgåede = self.morah.read_ou("8df2c488-8038-4b00-a600-000001460003")
        self.assertFalse("error_key" in udgåede)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
import logging
import typing
import urllib.error
from pprint import pprint

import prometheus_client
from prometheus_client import CollectorRegistry
from prometheus_client import Gauge
from ra_utils.async_to_sync import async_to_sync

from .gql_lora_cache_async import GQLLoraCache
from .gql_lora_cache_async import GqlLoraCacheSettings
from .old_lora_cache import OldLoraCache as LoraCache

logger = logging.getLogger(__name__)


# This is an equivalency test, an integration test, and is thus not designed to
# be run as part of the ci-pipeline
class TestEquivalence:
    def __init__(self, settings):
        self.settings: GqlLoraCacheSettings = settings

    # The old lora cache had some bugs, this accounts for them when comparing.
    # History and end dates for units, it connections, and managers were bugged, and
    # so was some of the addresses
    def account_for_know_errors(
        self, old_cache: dict, new_cache: dict, cache_type: str
    ) -> dict:

        ref_cache = {}

        lc_gql = GQLLoraCache(
            resolve_dar=True, full_history=True, skip_past=False, settings=self.settings
        )

        if cache_type != "addresses":
            ref_cache = self.new_cache_helper(lc_gql, cache_type)

        if cache_type == "addresses":
            for k, ocl in old_cache.items():
                ncl = new_cache[k]
                for i in range(len(ocl)):
                    if ocl[i]["scope"] == "DAR" and ocl[i]["value"] is None:
                        ocl[i]["value"] = ncl[i]["value"]

            return old_cache

        modified_old_cache = {}

        for k, old_list in old_cache.items():
            #
            if k not in new_cache:
                continue

            new_list: typing.List[typing.Dict] = []
            if k not in ref_cache:
                pprint(ref_cache)
            ref_list = ref_cache[k]

            new_cache_list = new_cache[k]
            for old in old_list:
                if old in new_cache_list:
                    new_list.append(old)
                    continue
                for ref in ref_list:
                    old["to_date"] = ref["to_date"]
                    old["from_date"] = ref["from_date"]
                    if old in new_cache_list and old not in new_list:
                        new_list.append(old)

            if not new_list:
                continue

            modified_old_cache[k] = new_list

        return modified_old_cache

    # Used as an entry generate a single cache, and get it
    @async_to_sync
    async def new_cache_helper(self, lc, cache_name):
        async with lc._setup_gql_client() as session:
            lc.gql_client_session = session
            match cache_name:
                case "facets":
                    await lc._cache_lora_facets()
                    return lc.facets
                case "classes":
                    await lc._cache_lora_classes()
                    return lc.classes
                case "it_systems":
                    await lc._cache_lora_itsystems()
                    return lc.itsystems
                case "users":
                    await lc._cache_lora_users()
                    return lc.users
                case "units":
                    await lc._cache_lora_units()
                    return lc.units
                case "engagements":
                    await lc._cache_lora_engagements()
                    return lc.facets
                case "roles":
                    await lc._cache_lora_roles()
                    return lc.roles
                case "leaves":
                    await lc._cache_lora_leaves()
                    return lc.leaves
                case "it_connections":
                    await lc._cache_lora_it_connections()
                    return lc.it_connections
                case "kles":
                    await lc._cache_lora_kles()
                    return lc.kles
                case "related":
                    await lc._cache_lora_related()
                    return lc.related
                case "managers":
                    await lc._cache_lora_managers()
                    return lc.managers
                case "associations":
                    await lc._cache_lora_associations()
                    return lc.associations
                case "addresses":
                    await lc._cache_lora_address()
                    return lc.addresses

    # Test populate, making a complete task. Also used for performance testing
    def test_populate(self, historic: bool, skip_past: bool, resolve_dar: bool):

        old_cache = LoraCache(
            resolve_dar=resolve_dar,
            full_history=historic,
            skip_past=skip_past,
            settings=self.settings.to_old_settings(),
        )

        new_cache = GQLLoraCache(
            resolve_dar=resolve_dar,
            full_history=historic,
            skip_past=skip_past,
            settings=self.settings,
        )

        new_cache.populate_cache(dry_run=False, skip_associations=False)

        old_cache.populate_cache(dry_run=False, skip_associations=False)
        old_cache.calculate_derived_unit_data()

        assert old_cache.facets == new_cache.facets
        assert old_cache.classes == new_cache.classes
        assert old_cache.itsystems == new_cache.itsystems
        assert old_cache.users == new_cache.users
        assert (
            self.account_for_know_errors(old_cache.units, new_cache.units, "units")
            == new_cache.units
        )

        assert old_cache.engagements == new_cache.engagements
        assert old_cache.roles == new_cache.roles
        assert old_cache.leaves == new_cache.leaves
        assert (
            self.account_for_know_errors(
                old_cache.it_connections, new_cache.it_connections, "it_connections"
            )
            == new_cache.it_connections
        )
        assert old_cache.kles == new_cache.kles
        assert old_cache.related == new_cache.related
        assert (
            self.account_for_know_errors(
                old_cache.managers,
                new_cache.managers,
                "managers",
            )
            == new_cache.managers
        )
        assert old_cache.associations == new_cache.associations
        assert (
            self.account_for_know_errors(
                old_cache.addresses, new_cache.addresses, "addresses"
            )
            == new_cache.addresses
        )


def notify_prometheus(
    settings: GqlLoraCacheSettings,
    job: str,
    start: bool = False,
    error: bool = False,
) -> None:
    """Used to send metrics to Prometheus

    Args:
    """
    registry = CollectorRegistry()
    name = "mo_end_time"
    if start:
        name = "mo_start_time"

    g_time = Gauge(
        name=name, documentation="Unixtime for job end time", registry=registry
    )
    g_time.set_to_current_time()

    g_ret_code = Gauge(
        name="mo_return_code",
        documentation="Return code of job",
        registry=registry,
    )
    if not error:
        g_ret_code.set(0)
    else:
        g_ret_code.inc(1)

    try:
        prometheus_client.exposition.pushadd_to_gateway(
            gateway=f"{settings.prometheus_pushgateway}:9091",
            job=job,
            registry=registry,
        )
    except urllib.error.URLError as ue:
        logger.warning("Cannot connect to Prometheus")
        logger.warning(ue)


def cache_equivalence():
    settings = GqlLoraCacheSettings()
    settings.start_logging_based_on_settings()

    job_name = "cache_equivalence_actual_state"
    try:
        notify_prometheus(settings=settings, job=job_name, start=True)
        tester = TestEquivalence(settings)
        tester.test_populate(historic=False, skip_past=True, resolve_dar=True)
        notify_prometheus(settings=settings, job=job_name)
    except Exception as e:
        notify_prometheus(settings=settings, job=job_name, error=True)
        raise e

    job_name = "cache_equivalence_historic"
    try:
        notify_prometheus(settings=settings, job=job_name, start=True)
        tester = TestEquivalence(settings)
        tester.test_populate(historic=True, skip_past=False, resolve_dar=True)
        notify_prometheus(settings=settings, job=job_name)
    except Exception as e:
        notify_prometheus(settings=settings, job=job_name, error=True)
        raise e

    job_name = "cache_equivalence_historic_skip_past"
    try:
        notify_prometheus(settings=settings, job=job_name, start=True)
        tester = TestEquivalence(settings)
        tester.test_populate(historic=True, skip_past=True, resolve_dar=True)
        notify_prometheus(settings=settings, job=job_name)
    except Exception as e:
        notify_prometheus(settings=settings, job=job_name, error=True)
        raise e
